from __future__ import annotations

import dataclasses
import enum
import functools
import keyword
import logging
import typing
from typing import Final, Self

import zigpy.types as t
from zigpy.typing import UNDEFINED, UndefinedType

_LOGGER = logging.getLogger(__name__)


def _hex_uint16_repr(v: int) -> str:
    return t.uint16_t(v)._hex_repr()


def ensure_valid_name(name: str | None) -> None:
    """Ensures that the name of an attribute or command is valid."""
    if name is not None and not name.isidentifier():
        raise ValueError(f"{name!r} is not a valid identifier name.")


class Status(t.enum8):
    SUCCESS = 0x00  # Operation was successful.
    FAILURE = 0x01  # Operation was not successful
    NOT_AUTHORIZED = 0x7E  # The sender of the command does not have
    RESERVED_FIELD_NOT_ZERO = 0x7F  # A reserved field/subfield/bit contains a
    MALFORMED_COMMAND = 0x80  # The command appears to contain the wrong
    UNSUP_CLUSTER_COMMAND = 0x81  # The specified cluster command is not
    UNSUP_GENERAL_COMMAND = 0x82  # The specified general ZCL command is not
    UNSUP_MANUF_CLUSTER_COMMAND = 0x83  # A manufacturer specific unicast,
    UNSUP_MANUF_GENERAL_COMMAND = 0x84  # A manufacturer specific unicast, ZCL
    INVALID_FIELD = 0x85  # At least one field of the command contains an
    UNSUPPORTED_ATTRIBUTE = 0x86  # The specified attribute does not exist on
    INVALID_VALUE = 0x87  # Out of range error, or set to a reserved value.
    READ_ONLY = 0x88  # Attempt to write a read only attribute.
    INSUFFICIENT_SPACE = 0x89  # An operation (e.g. an attempt to create an
    DUPLICATE_EXISTS = 0x8A  # An attempt to create an entry in a table failed
    NOT_FOUND = 0x8B  # The requested information (e.g. table entry)
    UNREPORTABLE_ATTRIBUTE = 0x8C  # Periodic reports cannot be issued for this
    INVALID_DATA_TYPE = 0x8D  # The data type given for an attribute is
    INVALID_SELECTOR = 0x8E  # The selector for an attribute is incorrect.
    WRITE_ONLY = 0x8F  # A request has been made to read an attribute
    INCONSISTENT_STARTUP_STATE = 0x90  # Setting the requested values would put
    DEFINED_OUT_OF_BAND = 0x91  # An attempt has been made to write an
    INCONSISTENT = (
        0x92  # The supplied values (e.g., contents of table cells) are inconsistent
    )
    ACTION_DENIED = 0x93  # The credentials presented by the device sending the
    TIMEOUT = 0x94  # The exchange was aborted due to excessive response time
    ABORT = 0x95  # Failed case when a client or a server decides to abort the upgrade process
    INVALID_IMAGE = 0x96  # Invalid OTA upgrade image (ex. failed signature
    WAIT_FOR_DATA = 0x97  # Server does not have data block available yet
    NO_IMAGE_AVAILABLE = 0x98  # No OTA upgrade image available for a particular client
    REQUIRE_MORE_IMAGE = 0x99  # The client still requires more OTA upgrade image
    NOTIFICATION_PENDING = 0x9A  # The command has been received and is being processed
    HARDWARE_FAILURE = 0xC0  # An operation was unsuccessful due to a
    SOFTWARE_FAILURE = 0xC1  # An operation was unsuccessful due to a
    CALIBRATION_ERROR = 0xC2  # An error occurred during calibration
    UNSUPPORTED_CLUSTER = 0xC3  # The cluster is not supported

    @classmethod
    def _missing_(cls, value):
        chained = t.APSStatus(value)
        status = cls._member_type_.__new__(cls, chained.value)
        status._name_ = chained.name
        status._value_ = value
        return status


class DataClass(enum.Enum):
    Null = 0
    Analog = 1
    Discrete = 2
    Composite = 3


# TODO: Backwards compatibility, remove later
Null = DataClass.Null
Analog = DataClass.Analog
Discrete = DataClass.Discrete
Composite = DataClass.Composite


class Unknown(t.NoData):
    pass


@dataclasses.dataclass()
class TypeValue:
    type: t.uint8_t = dataclasses.field(default=None)
    value: typing.Any = dataclasses.field(default=None)

    def __init__(self, type: t.uint8_t | None = None, value: typing.Any = None) -> None:
        # "Copy constructor"
        if type is not None and value is None and isinstance(type, self.__class__):
            other = type
            type = other.type  # noqa: A001
            value = other.value

        self.type = type
        self.value = value

    def serialize(self) -> bytes:
        return self.type.to_bytes(1, "little") + self.value.serialize()

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[TypeValue, bytes]:
        data_type, data = t.uint8_t.deserialize(data)
        python_type = DataType.from_type_id(data_type).python_type
        value, data = python_type.deserialize(data)

        return cls(type=data_type, value=value), data

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"type={type(self.value).__name__}, value={self.value!r}"
            f")"
        )


class TypedCollection(TypeValue):
    @classmethod
    def deserialize(cls, data):
        data_type, data = t.uint8_t.deserialize(data)
        python_type = DataType.from_type_id(data_type).python_type
        values, data = t.LVList[python_type, t.uint16_t].deserialize(data)

        return cls(type=data_type, value=values), data


class Array(TypedCollection):
    pass


class Bag(TypedCollection):
    pass


class Set(TypedCollection):
    pass  # ToDo: Make this a real set?


class ZCLStructure(t.LVList, item_type=TypeValue, length_type=t.uint16_t):
    """ZCL Structure data type."""


class DataTypeId(t.enum8):
    unk = 0xFF
    nodata = 0x00
    data8 = 0x08
    data16 = 0x09
    data24 = 0x0A
    data32 = 0x0B
    data40 = 0x0C
    data48 = 0x0D
    data56 = 0x0E
    data64 = 0x0F
    bool_ = 0x10
    map8 = 0x18
    map16 = 0x19
    map24 = 0x1A
    map32 = 0x1B
    map40 = 0x1C
    map48 = 0x1D
    map56 = 0x1E
    map64 = 0x1F
    uint8 = 0x20
    uint16 = 0x21
    uint24 = 0x22
    uint32 = 0x23
    uint40 = 0x24
    uint48 = 0x25
    uint56 = 0x26
    uint64 = 0x27
    int8 = 0x28
    int16 = 0x29
    int24 = 0x2A
    int32 = 0x2B
    int40 = 0x2C
    int48 = 0x2D
    int56 = 0x2E
    int64 = 0x2F
    enum8 = 0x30
    enum16 = 0x31
    semi = 0x38
    single = 0x39
    double = 0x3A
    octstr = 0x41
    string = 0x42
    octstr16 = 0x43
    string16 = 0x44
    array = 0x48
    struct = 0x4C
    set = 0x50
    bag = 0x51
    ToD = 0xE0
    date = 0xE1
    UTC = 0xE2
    clusterId = 0xE8  # noqa: N815
    attribId = 0xE9  # noqa: N815
    bacOID = 0xEA  # noqa: N815
    EUI64 = 0xF0
    key128 = 0xF1


@dataclasses.dataclass(frozen=True)
class DataTypeInfo:
    type_id: DataTypeId
    python_type: type
    type_class: DataClass
    description: str
    non_value: typing.Any | None


class DataType(DataTypeInfo, enum.Enum):
    unk = (
        DataTypeId.unk,
        Unknown,
        DataClass.Null,
        "Unknown",
        None,
    )
    nodata = (
        DataTypeId.nodata,
        t.NoData,
        DataClass.Null,
        "No data",
        None,
    )
    data8 = (
        DataTypeId.data8,
        t.data8,
        DataClass.Discrete,
        "General",
        None,
    )
    data16 = (
        DataTypeId.data16,
        t.data16,
        DataClass.Discrete,
        "General",
        None,
    )
    data24 = (
        DataTypeId.data24,
        t.data24,
        DataClass.Discrete,
        "General",
        None,
    )
    data32 = (
        DataTypeId.data32,
        t.data32,
        DataClass.Discrete,
        "General",
        None,
    )
    data40 = (
        DataTypeId.data40,
        t.data40,
        DataClass.Discrete,
        "General",
        None,
    )
    data48 = (
        DataTypeId.data48,
        t.data48,
        DataClass.Discrete,
        "General",
        None,
    )
    data56 = (
        DataTypeId.data56,
        t.data56,
        DataClass.Discrete,
        "General",
        None,
    )
    data64 = (
        DataTypeId.data64,
        t.data64,
        DataClass.Discrete,
        "General",
        None,
    )
    bool_ = (
        DataTypeId.bool_,
        t.Bool,
        DataClass.Discrete,
        "Boolean",
        t.Bool(0xFF),
    )
    map8 = (
        DataTypeId.map8,
        t.bitmap8,
        DataClass.Discrete,
        "Bitmap",
        None,
    )
    map16 = (
        DataTypeId.map16,
        t.bitmap16,
        DataClass.Discrete,
        "Bitmap",
        None,
    )
    map24 = (
        DataTypeId.map24,
        t.bitmap24,
        DataClass.Discrete,
        "Bitmap",
        None,
    )
    map32 = (
        DataTypeId.map32,
        t.bitmap32,
        DataClass.Discrete,
        "Bitmap",
        None,
    )
    map40 = (
        DataTypeId.map40,
        t.bitmap40,
        DataClass.Discrete,
        "Bitmap",
        None,
    )
    map48 = (
        DataTypeId.map48,
        t.bitmap48,
        DataClass.Discrete,
        "Bitmap",
        None,
    )
    map56 = (
        DataTypeId.map56,
        t.bitmap56,
        DataClass.Discrete,
        "Bitmap",
        None,
    )
    map64 = (
        DataTypeId.map64,
        t.bitmap64,
        DataClass.Discrete,
        "Bitmap",
        None,
    )
    uint8 = (
        DataTypeId.uint8,
        t.uint8_t,
        DataClass.Analog,
        "Unsigned 8-bit integer",
        t.uint8_t(0xFF),
    )
    uint16 = (
        DataTypeId.uint16,
        t.uint16_t,
        DataClass.Analog,
        "Unsigned 16-bit integer",
        t.uint16_t(0xFFFF),
    )
    uint24 = (
        DataTypeId.uint24,
        t.uint24_t,
        DataClass.Analog,
        "Unsigned 24-bit integer",
        t.uint24_t(0xFFFFFF),
    )
    uint32 = (
        DataTypeId.uint32,
        t.uint32_t,
        DataClass.Analog,
        "Unsigned 32-bit integer",
        t.uint32_t(0xFFFFFFFF),
    )
    uint40 = (
        DataTypeId.uint40,
        t.uint40_t,
        DataClass.Analog,
        "Unsigned 40-bit integer",
        t.uint40_t(0xFFFFFFFFFF),
    )
    uint48 = (
        DataTypeId.uint48,
        t.uint48_t,
        DataClass.Analog,
        "Unsigned 48-bit integer",
        t.uint48_t(0xFFFFFFFFFFFF),
    )
    uint56 = (
        DataTypeId.uint56,
        t.uint56_t,
        DataClass.Analog,
        "Unsigned 56-bit integer",
        t.uint56_t(0xFFFFFFFFFFFFFF),
    )
    uint64 = (
        DataTypeId.uint64,
        t.uint64_t,
        DataClass.Analog,
        "Unsigned 64-bit integer",
        t.uint64_t(0xFFFFFFFFFFFFFF),
    )
    int8 = (
        DataTypeId.int8,
        t.int8s,
        DataClass.Analog,
        "Signed 8-bit integer",
        t.int8s(-0x80),
    )
    int16 = (
        DataTypeId.int16,
        t.int16s,
        DataClass.Analog,
        "Signed 16-bit integer",
        t.int16s(-0x8000),
    )
    int24 = (
        DataTypeId.int24,
        t.int24s,
        DataClass.Analog,
        "Signed 24-bit integer",
        t.int24s(-0x800000),
    )
    int32 = (
        DataTypeId.int32,
        t.int32s,
        DataClass.Analog,
        "Signed 32-bit integer",
        t.int32s(-0x80000000),
    )
    int40 = (
        DataTypeId.int40,
        t.int40s,
        DataClass.Analog,
        "Signed 40-bit integer",
        t.int40s(-0x8000000000),
    )
    int48 = (
        DataTypeId.int48,
        t.int48s,
        DataClass.Analog,
        "Signed 48-bit integer",
        t.int48s(-0x800000000000),
    )
    int56 = (
        DataTypeId.int56,
        t.int56s,
        DataClass.Analog,
        "Signed 56-bit integer",
        t.int56s(-0x80000000000000),
    )
    int64 = (
        DataTypeId.int64,
        t.int64s,
        DataClass.Analog,
        "Signed 64-bit integer",
        t.int64s(-0x80000000000000),
    )
    enum8 = (
        DataTypeId.enum8,
        t.enum8,
        DataClass.Discrete,
        "8-bit enumeration",
        t.enum8(0xFF),
    )
    enum16 = (
        DataTypeId.enum16,
        t.enum16,
        DataClass.Discrete,
        "16-bit enumeration",
        t.enum16(0xFF),
    )
    semi = (
        DataTypeId.semi,
        t.Half,
        DataClass.Analog,
        "Semi-precision",
        t.Half(float("nan")),
    )
    single = (
        DataTypeId.single,
        t.Single,
        DataClass.Analog,
        "Single precision",
        t.Single(float("nan")),
    )
    double = (
        DataTypeId.double,
        t.Double,
        DataClass.Analog,
        "Double precision",
        t.Double(float("nan")),
    )
    octstr = (
        DataTypeId.octstr,
        t.LVBytes,
        DataClass.Discrete,
        "Octet string",
        None,
    )
    string = (
        DataTypeId.string,
        t.CharacterString,
        DataClass.Discrete,
        "Character string",
        None,
    )
    octstr16 = (
        DataTypeId.octstr16,
        t.LongOctetString,
        DataClass.Discrete,
        "Long octet string",
        None,
    )
    string16 = (
        DataTypeId.string16,
        t.LongCharacterString,
        DataClass.Discrete,
        "Long character string",
        None,
    )
    array = (
        DataTypeId.array,
        Array,
        DataClass.Discrete,
        "Array",
        None,
    )
    struct = (
        DataTypeId.struct,
        ZCLStructure,
        DataClass.Discrete,
        "Structure",
        None,
    )
    set = (
        DataTypeId.set,
        Set,
        DataClass.Discrete,
        "Set",
        None,
    )
    bag = (
        DataTypeId.bag,
        Bag,
        DataClass.Discrete,
        "Bag",
        None,
    )
    ToD = (
        DataTypeId.ToD,
        t.TimeOfDay,
        DataClass.Analog,
        "Time of day",
        t.TimeOfDay(hours=0xFF, minutes=0xFF, seconds=0xFF, hundredths=0xFF),
    )
    date = (
        DataTypeId.date,
        t.Date,
        DataClass.Analog,
        "Date",
        t.Date(years_since_1900=0xFF, month=0xFF, day=0xFF, day_of_week=0xFF),
    )
    UTC = (
        DataTypeId.UTC,
        t.UTCTime,
        DataClass.Analog,
        "UTCTime",
        t.UTCTime(0xFFFFFFFF),
    )
    clusterId = (  # noqa: N815
        DataTypeId.clusterId,
        t.ClusterId,
        DataClass.Discrete,
        "Cluster ID",
        t.ClusterId(0xFFFF),
    )
    attribId = (  # noqa: N815
        DataTypeId.attribId,
        t.AttributeId,
        DataClass.Discrete,
        "Attribute ID",
        t.AttributeId(0xFFFF),
    )
    bacOID = (  # noqa: N815
        DataTypeId.bacOID,
        t.BACNetOid,
        DataClass.Discrete,
        "BACNet OID",
        t.BACNetOid(0xFFFFFFFF),
    )
    EUI64 = (
        DataTypeId.EUI64,
        t.EUI64,
        DataClass.Discrete,
        "IEEE address",
        t.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"),
    )
    key128 = (
        DataTypeId.key128,
        t.KeyData,
        DataClass.Discrete,
        "128-bit security key",
        t.KeyData.convert("FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF"),
    )

    @classmethod
    @functools.cache
    def _python_type_index(cls: type[Self]) -> dict[type, Self]:
        return {d.python_type: d for d in cls}

    @classmethod
    def from_python_type(cls: type[Self], python_type: type) -> Self:
        """Return Zigbee Datatype ID for a give python type."""
        python_type_index = cls._python_type_index()

        # We return the most specific parent class
        for parent_cls in python_type.__mro__:
            if parent_cls in python_type_index:
                return python_type_index[parent_cls]

        return cls.unk

    @classmethod
    @functools.cache
    def _data_type_index(cls: type[Self]) -> dict[DataTypeId, Self]:
        return {d.type_id: d for d in cls}

    @classmethod
    def from_type_id(cls: type[Self], type_id: DataTypeId) -> Self:
        return cls._data_type_index()[type_id]


@dataclasses.dataclass()
class ReadAttributeRecord:
    """Read Attribute Record."""

    attrid: t.uint16_t
    status: Status
    value: TypeValue | Array | Bag | Set | None

    def __init__(
        self,
        attrid: t.uint16_t | Self = t.uint16_t(0x0000),
        status: Status = Status.SUCCESS,
        value: TypeValue | Array | Bag | Set | None = None,
    ) -> None:
        if isinstance(attrid, self.__class__):
            # "Copy constructor"
            self.attrid = attrid.attrid
            self.status = attrid.status
            self.value = attrid.value
            return

        self.attrid = t.uint16_t(attrid)
        self.status = Status(status)
        self.value = value

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[Self, bytes]:
        attrid, data = t.uint16_t.deserialize(data)
        status, data = Status.deserialize(data)
        value = None

        if status == Status.SUCCESS:
            type_id, data = DataTypeId.deserialize(data)

            # Arrays, Sets, and Bags are treated differently
            if type_id in (DataTypeId.array, DataTypeId.set, DataTypeId.bag):
                value, data = DataType.from_type_id(type_id).python_type.deserialize(
                    data
                )
            else:
                value, data = TypeValue.deserialize(type_id.serialize() + data)

        return cls(attrid=attrid, status=status, value=value), data

    def serialize(self) -> bytes:
        data = self.attrid.serialize()
        data += self.status.serialize()

        if self.status == Status.SUCCESS:
            assert self.value is not None

            if isinstance(self.value, Array | Set | Bag):
                data += (
                    DataType.from_python_type(type(self.value)).type_id.serialize()
                    + self.value.serialize()
                )
            else:
                data += self.value.serialize()

        return data


class Attribute(t.Struct):
    attrid: t.uint16_t = t.StructField(repr=_hex_uint16_repr)
    value: TypeValue


class WriteAttributesStatusRecord(t.Struct):
    status: Status
    attrid: t.uint16_t = t.StructField(
        requires=lambda s: s.status != Status.SUCCESS, repr=_hex_uint16_repr
    )


class WriteAttributesResponse(list):
    """Write Attributes response list.

    Response to Write Attributes request should contain only success status, in
    case when all attributes were successfully written or list of status + attr_id
    records for all failed writes.
    """

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[WriteAttributesResponse, bytes]:
        record, data = WriteAttributesStatusRecord.deserialize(data)
        r = cls([record])
        if record.status == Status.SUCCESS:
            return r, data

        while len(data) >= 3:
            record, data = WriteAttributesStatusRecord.deserialize(data)
            r.append(record)
        return r, data

    def serialize(self):
        failed = [record for record in self if record.status != Status.SUCCESS]
        if failed:
            return b"".join(
                [WriteAttributesStatusRecord(i).serialize() for i in failed]
            )
        return Status.SUCCESS.serialize()


class ReportingDirection(t.enum8):
    SendReports = 0x00
    ReceiveReports = 0x01


class AttributeReportingStatus(t.enum8):
    Pending = 0x00
    Attribute_Reporting_Complete = 0x01


class AttributeReportingConfig:
    def __init__(self, other: AttributeReportingConfig | None = None) -> None:
        if isinstance(other, self.__class__):
            self.direction: ReportingDirection = other.direction
            self.attrid: t.uint16_t = other.attrid
            if self.direction == ReportingDirection.ReceiveReports:
                self.timeout: int = other.timeout
                return
            self.datatype: DataTypeId = other.datatype
            self.min_interval: int = other.min_interval
            self.max_interval: int = other.max_interval
            self.reportable_change: int = other.reportable_change

    def serialize(self, *, _only_dir_and_attrid: bool = False) -> bytes:
        r = ReportingDirection(self.direction).serialize()
        r += t.uint16_t(self.attrid).serialize()

        if _only_dir_and_attrid:
            return r

        if self.direction == ReportingDirection.ReceiveReports:
            r += t.uint16_t(self.timeout).serialize()
        else:
            r += t.uint8_t(self.datatype).serialize()
            r += t.uint16_t(self.min_interval).serialize()
            r += t.uint16_t(self.max_interval).serialize()

            try:
                data_type = DataType.from_type_id(self.datatype)
            except KeyError:
                _LOGGER.warning(
                    "Unknown ZCL type %d, not setting reportable change", self.datatype
                )
            else:
                if data_type.type_class is Analog:
                    r += data_type.python_type(self.reportable_change).serialize()

        return r

    @classmethod
    def deserialize(
        cls, data, *, _only_dir_and_attrid: bool = False
    ) -> tuple[AttributeReportingConfig, bytes]:
        self = cls()
        self.direction, data = ReportingDirection.deserialize(data)
        self.attrid, data = t.uint16_t.deserialize(data)

        # The report is only a direction and attribute
        if _only_dir_and_attrid:
            return self, data

        if self.direction == ReportingDirection.ReceiveReports:
            # Requesting things to be received by me
            self.timeout, data = t.uint16_t.deserialize(data)
        else:
            # Notifying that I will report things to you
            self.datatype, data = t.uint8_t.deserialize(data)
            self.min_interval, data = t.uint16_t.deserialize(data)
            self.max_interval, data = t.uint16_t.deserialize(data)

            try:
                data_type = DataType.from_type_id(self.datatype)
            except KeyError:
                _LOGGER.warning(
                    "Unknown ZCL type %d, cannot read reportable change", self.datatype
                )
            else:
                if data_type.type_class is Analog:
                    self.reportable_change, data = data_type.python_type.deserialize(
                        data
                    )

        return self, data

    def __repr__(self) -> str:
        r = f"{self.__class__.__name__}("
        r += f"direction={self.direction}"
        r += f", attrid=0x{self.attrid:04X}"

        if self.direction == ReportingDirection.ReceiveReports:
            r += f", timeout={self.timeout}"
        elif hasattr(self, "datatype"):
            r += f", datatype={self.datatype}"
            r += f", min_interval={self.min_interval}"
            r += f", max_interval={self.max_interval}"

            if self.reportable_change is not None:
                r += f", reportable_change={self.reportable_change}"

        r += ")"

        return r


class AttributeReportingConfigWithStatus(t.Struct):
    status: Status
    config: AttributeReportingConfig

    @classmethod
    def deserialize(
        cls, data: bytes
    ) -> tuple[AttributeReportingConfigWithStatus, bytes]:
        status, data = Status.deserialize(data)

        # FIXME: The reporting configuration will not include anything other than the
        # direction and the attribute ID when the status is not successful. This
        # information isn't a part of the attribute reporting config structure so we
        # have to pass it in externally.
        config, data = AttributeReportingConfig.deserialize(
            data, _only_dir_and_attrid=(status != Status.SUCCESS)
        )

        return cls(status=status, config=config), data

    def serialize(self) -> bytes:
        return self.status.serialize() + self.config.serialize(
            _only_dir_and_attrid=(self.status != Status.SUCCESS)
        )


class ConfigureReportingResponseRecord(t.Struct):
    status: Status
    direction: ReportingDirection
    attrid: t.uint16_t = t.StructField(repr=_hex_uint16_repr)

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[ConfigureReportingResponseRecord, bytes]:
        r = cls()
        r.status, data = Status.deserialize(data)
        if r.status == Status.SUCCESS:
            r.direction, data = t.Optional(t.uint8_t).deserialize(data)
            if r.direction is not None:
                r.direction = ReportingDirection(r.direction)
            r.attrid, data = t.Optional(t.uint16_t).deserialize(data)
            return r, data

        r.direction, data = ReportingDirection.deserialize(data)
        r.attrid, data = t.uint16_t.deserialize(data)
        return r, data

    def serialize(self):
        r = Status(self.status).serialize()
        if self.status != Status.SUCCESS:
            r += ReportingDirection(self.direction).serialize()
            r += t.uint16_t(self.attrid).serialize()
        return r

    def __repr__(self) -> str:
        r = f"{self.__class__.__name__}(status={self.status}"
        if self.status != Status.SUCCESS:
            r += f", direction={self.direction}, attrid={self.attrid}"
        r += ")"
        return r


class ConfigureReportingResponse(t.List[ConfigureReportingResponseRecord]):
    # In the case of successful configuration of all attributes, only a single
    # attribute status record SHALL be included in the command, with the status
    # field set to SUCCESS and the direction and attribute identifier fields omitted

    def serialize(self):
        if not self:
            raise ValueError("Cannot serialize empty list")

        failed = [record for record in self if record.status != Status.SUCCESS]

        if not failed:
            return ConfigureReportingResponseRecord(status=Status.SUCCESS).serialize()

        # Note that attribute status records are not included for successfully
        # configured attributes, in order to save bandwidth.
        return b"".join(
            [ConfigureReportingResponseRecord(r).serialize() for r in failed]
        )


class ReadReportingConfigRecord(t.Struct):
    direction: t.uint8_t
    attrid: t.uint16_t


class DiscoverAttributesResponseRecord(t.Struct):
    attrid: t.uint16_t
    datatype: t.uint8_t


class AttributeAccessControl(t.bitmap8):
    READ = 0x01
    WRITE = 0x02
    REPORT = 0x04


class DiscoverAttributesExtendedResponseRecord(t.Struct):
    attrid: t.uint16_t
    datatype: t.uint8_t
    acl: AttributeAccessControl


class SelectorOperation(t.enum4):
    Write = 0x0
    Add = 0x1
    Remove = 0x2


class Selector(t.Struct):
    """ZCL attribute selector for structured read/write commands."""

    depth: t.uint4_t
    op: SelectorOperation = t.StructField(default=SelectorOperation.Write)
    indexes: t.List[t.uint16_t] = t.StructField(length=lambda s: s.depth)


class ReadAttributeStructured(t.Struct):
    attrid: t.uint16_t
    selector: Selector


class WriteAttributeStructured(t.Struct):
    attrid: t.uint16_t
    selector: Selector
    value: TypeValue


class WriteAttributesStructuredStatusRecord(t.Struct):
    status: Status
    attrid: t.uint16_t = t.StructField(
        requires=lambda s: s.status != Status.SUCCESS, repr=_hex_uint16_repr
    )
    selector: Selector = t.StructField(requires=lambda s: s.status != Status.SUCCESS)


class WriteAttributesStructuredResponse(list):
    """Write Attributes Structured response list.

    Response to Write Attributes Structured request should contain only success status,
    in case when all attributes were successfully written or list of status + attr_id
    records for all failed writes.
    """

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[Self, bytes]:
        record, data = WriteAttributesStructuredStatusRecord.deserialize(data)
        r = cls([record])
        if record.status == Status.SUCCESS:
            return r, data

        while len(data) >= 4:
            record, data = WriteAttributesStructuredStatusRecord.deserialize(data)
            r.append(record)
        return r, data

    def serialize(self) -> bytes:
        failed = [record for record in self if record.status != Status.SUCCESS]
        if failed:
            return b"".join(
                [WriteAttributesStructuredStatusRecord(i).serialize() for i in failed]
            )
        return Status.SUCCESS.serialize()


class FrameType(t.enum2):
    """ZCL Frame Type."""

    GLOBAL_COMMAND = 0b00
    CLUSTER_COMMAND = 0b01
    RESERVED_2 = 0b10
    RESERVED_3 = 0b11


class Direction(t.enum1):
    """ZCL frame control direction."""

    Client_to_Server = 0
    Server_to_Client = 1

    @classmethod
    def _from_is_reply(cls, is_reply: bool) -> Direction:
        return cls.Server_to_Client if is_reply else cls.Client_to_Server

    def flip(self) -> Self:
        return (
            self.Server_to_Client
            if self == self.Client_to_Server
            else self.Client_to_Server
        )


class FrameControl(t.IntStruct, t.uint8_t):
    """The frame control field contains information defining the command type
    and other control flags.
    """

    frame_type: FrameType
    is_manufacturer_specific: t.uint1_t
    direction: Direction
    disable_default_response: t.uint1_t
    reserved: t.uint3_t

    @classmethod
    def cluster(
        cls,
        direction: Direction = Direction.Client_to_Server,
        is_manufacturer_specific: bool = False,
    ):
        return cls(
            frame_type=FrameType.CLUSTER_COMMAND,
            is_manufacturer_specific=is_manufacturer_specific,
            direction=direction,
            disable_default_response=(direction == Direction.Server_to_Client),
            reserved=0b000,
        )

    @classmethod
    def general(
        cls,
        direction: Direction = Direction.Client_to_Server,
        is_manufacturer_specific: bool = False,
    ):
        return cls(
            frame_type=FrameType.GLOBAL_COMMAND,
            is_manufacturer_specific=is_manufacturer_specific,
            direction=direction,
            disable_default_response=(direction == Direction.Server_to_Client),
            reserved=0b000,
        )

    @property
    def is_cluster(self) -> bool:
        """Return True if command is a local cluster specific command."""
        return bool(self.frame_type == FrameType.CLUSTER_COMMAND)

    @property
    def is_general(self) -> bool:
        """Return True if command is a global ZCL command."""
        return bool(self.frame_type == FrameType.GLOBAL_COMMAND)


class ZCLHeader(t.Struct):
    # Kept for backwards compatibility
    NO_MANUFACTURER_ID: Final = None

    frame_control: FrameControl
    manufacturer: t.uint16_t = t.StructField(
        requires=lambda hdr: hdr.frame_control.is_manufacturer_specific
    )
    tsn: t.uint8_t
    command_id: t.uint8_t

    def __new__(
        cls: type[Self],
        frame_control: FrameControl | None = None,
        manufacturer: t.uint16_t | None = None,
        tsn: int | t.uint8_t | None = None,
        command_id: int | GeneralCommand | None = None,
    ) -> Self:
        if frame_control is not None and manufacturer is not None:
            frame_control = frame_control.replace(is_manufacturer_specific=True)

        return super().__new__(cls, frame_control, manufacturer, tsn, command_id)

    @property
    def direction(self) -> Direction:
        """Return direction of Frame Control."""
        return self.frame_control.direction

    def __setattr__(
        self,
        name: str,
        value: t.uint16_t | FrameControl | t.uint8_t | GeneralCommand | None,
    ) -> None:
        super().__setattr__(name, value)

        if name == "manufacturer" and self.frame_control is not None:
            self.frame_control = self.frame_control.replace(
                is_manufacturer_specific=value is not None
            )

    @classmethod
    def general(
        cls,
        tsn: int | t.uint8_t,
        command_id: int | t.uint8_t,
        manufacturer: int | t.uint16_t | None = None,
        direction: Direction = Direction.Client_to_Server,
    ) -> ZCLHeader:
        return cls(
            frame_control=FrameControl.general(
                direction=direction,
                is_manufacturer_specific=(manufacturer is not None),
            ),
            manufacturer=manufacturer,
            tsn=tsn,
            command_id=command_id,
        )

    @classmethod
    def cluster(
        cls,
        tsn: int | t.uint8_t,
        command_id: int | t.uint8_t,
        manufacturer: int | t.uint16_t | None = None,
        direction: Direction = Direction.Client_to_Server,
    ) -> ZCLHeader:
        return cls(
            frame_control=FrameControl.cluster(
                direction=direction,
                is_manufacturer_specific=(manufacturer is not None),
            ),
            manufacturer=manufacturer,
            tsn=tsn,
            command_id=command_id,
        )


@dataclasses.dataclass(frozen=True)
class ZCLCommandDef(t.BaseDataclassMixin):
    id: t.uint8_t = None
    schema: type[CommandSchema] = None
    direction: Direction = None
    is_manufacturer_specific: bool | None = None

    # set later
    name: str = None
    manufacturer_code: t.uint16_t | UndefinedType | None = UNDEFINED

    def __post_init__(self) -> None:
        # Backwards compatibility with positional syntax where the name was first
        if isinstance(self.id, str):
            object.__setattr__(self, "name", self.id)
            object.__setattr__(self, "id", None)

        ensure_valid_name(self.name)

        if isinstance(self.direction, bool):
            object.__setattr__(
                self, "direction", Direction._from_is_reply(self.direction)
            )

        if self.manufacturer_code is not UNDEFINED:
            object.__setattr__(
                self, "is_manufacturer_specific", self.manufacturer_code is not None
            )

    def with_compiled_schema(self) -> ZCLCommandDef:
        """Return a copy of the ZCL command definition object with its dictionary command
        schema converted into a `CommandSchema` subclass.
        """

        if isinstance(self.schema, tuple):
            raise ValueError(  # noqa: TRY004
                f"Tuple schemas are deprecated: {self.schema!r}. Use a dictionary or a"
                f" Struct subclass."
            )
        elif not isinstance(self.schema, dict):
            # If the schema is already a struct, do nothing
            self.schema.command = self
            return self

        assert self.id is not None
        assert self.name is not None

        cls_attrs = {
            "__annotations__": {},
            "command": self,
        }

        for name, param_type in self.schema.items():
            plain_name = name.rstrip("?")

            # Make sure parameters with names like "foo bar" and "class" can't exist
            if not plain_name.isidentifier() or keyword.iskeyword(plain_name):
                raise ValueError(
                    f"Schema parameter {name} must be a valid Python identifier"
                )

            cls_attrs["__annotations__"][plain_name] = "None"
            cls_attrs[plain_name] = t.StructField(
                type=param_type,
                optional=name.endswith("?"),
            )

        schema = type(self.name, (CommandSchema,), cls_attrs)

        return self.replace(schema=schema)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"id=0x{self.id:02X}, "
            f"name={self.name!r}, "
            f"direction={self.direction}, "
            f"schema={self.schema}, "
            f"is_manufacturer_specific={self.is_manufacturer_specific}, "
            f"manufacturer_code={self.manufacturer_code}"
            f")"
        )


class CommandSchema(t.Struct, tuple):  # noqa: SLOT001, PLW1641
    """Struct subclass that behaves more like a tuple."""

    command: ZCLCommandDef = None

    def __iter__(self):
        return iter(self.as_tuple())

    def __getitem__(
        self, item: slice | typing.SupportsIndex
    ) -> typing.Any | tuple[typing.Any, ...]:
        return self.as_tuple()[item]

    def __len__(self) -> int:
        return len(self.as_tuple())

    def __eq__(self, other) -> bool:
        if isinstance(other, tuple) and not isinstance(other, type(self)):
            return self.as_tuple() == other

        return super().__eq__(other)


class ReadAttributesResponse(CommandSchema):
    status_records: t.List[ReadAttributeRecord]


class WriteAttributesResponseSchema(CommandSchema):
    status_records: WriteAttributesResponse


class WriteAttributesStructuredResponseSchema(CommandSchema):
    status_records: WriteAttributesStructuredResponse


class DefaultResponse(CommandSchema):
    command_id: t.uint8_t
    status: Status


class ZCLAttributeAccess(enum.Flag):
    NONE = 0
    Read = 1
    Write = 2
    Write_Optional = 4
    Report = 8
    Scene = 16

    _names: dict[ZCLAttributeAccess, str]

    @classmethod
    @functools.lru_cache(None)
    def from_str(cls, value: str) -> Self:
        orig_value = value
        access = cls.NONE

        while value:
            for mode, prefix in cls._names.items():
                if value.startswith(prefix):
                    value = value[len(prefix) :]
                    access |= mode
                    break
            else:
                raise ValueError(f"Invalid access mode: {orig_value!r}")

        return cls(access)


ZCLAttributeAccess._names = {
    ZCLAttributeAccess.Write_Optional: "*w",
    ZCLAttributeAccess.Write: "w",
    ZCLAttributeAccess.Read: "r",
    ZCLAttributeAccess.Report: "p",
    ZCLAttributeAccess.Scene: "s",
}


@dataclasses.dataclass(frozen=True)
class ZCLAttributeDef(t.BaseDataclassMixin):
    id: t.uint16_t = None
    type: typing.Any = None
    zcl_type: DataTypeId = None
    access: ZCLAttributeAccess = (
        ZCLAttributeAccess.Read | ZCLAttributeAccess.Write | ZCLAttributeAccess.Report
    )
    mandatory: bool = False
    is_manufacturer_specific: bool | None = None

    # These are (optionally) computed later in the ZCL cluster subclass hook
    name: str = None
    manufacturer_code: t.uint16_t | UndefinedType | None = UNDEFINED

    def __post_init__(self) -> None:
        # Backwards compatibility with positional syntax where the name was first
        if isinstance(self.id, str):
            object.__setattr__(self, "name", self.id)
            object.__setattr__(self, "id", None)

        if self.id is not None and not isinstance(self.id, t.uint16_t):
            object.__setattr__(self, "id", t.uint16_t(self.id))

        if isinstance(self.access, str):
            object.__setattr__(self, "access", ZCLAttributeAccess.from_str(self.access))

        if self.zcl_type is None:
            object.__setattr__(
                self, "zcl_type", DataType.from_python_type(self.type).type_id
            )

        ensure_valid_name(self.name)

        if self.manufacturer_code is not UNDEFINED:
            object.__setattr__(
                self, "is_manufacturer_specific", self.manufacturer_code is not None
            )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"id=0x{self.id:04X}, "
            f"name={self.name!r}, "
            f"type={self.type}, "
            f"zcl_type={self.zcl_type}, "
            f"access={self.access!r}, "
            f"mandatory={self.mandatory!r}, "
            f"is_manufacturer_specific={self.is_manufacturer_specific}, "
            f"manufacturer_code={self.manufacturer_code}"
            f")"
        )


class IterableMemberMeta(type):
    def __iter__(cls) -> typing.Iterator[typing.Any]:
        for name in dir(cls):
            if name.startswith("_"):
                continue

            value = getattr(cls, name)
            if value is None:
                continue

            yield value


class BaseCommandDefs(metaclass=IterableMemberMeta):
    pass


class BaseAttributeDefs(metaclass=IterableMemberMeta):
    pass


class GeneralCommand(t.enum8):
    """ZCL Foundation General Command IDs."""

    Read_Attributes = 0x00
    Read_Attributes_rsp = 0x01
    Write_Attributes = 0x02
    Write_Attributes_Undivided = 0x03
    Write_Attributes_rsp = 0x04
    Write_Attributes_No_Response = 0x05
    Configure_Reporting = 0x06
    Configure_Reporting_rsp = 0x07
    Read_Reporting_Configuration = 0x08
    Read_Reporting_Configuration_rsp = 0x09
    Report_Attributes = 0x0A
    Default_Response = 0x0B
    Discover_Attributes = 0x0C
    Discover_Attributes_rsp = 0x0D
    Read_Attributes_Structured = 0x0E
    Write_Attributes_Structured = 0x0F
    Write_Attributes_Structured_rsp = 0x10
    Discover_Commands_Received = 0x11
    Discover_Commands_Received_rsp = 0x12
    Discover_Commands_Generated = 0x13
    Discover_Commands_Generated_rsp = 0x14
    Discover_Attribute_Extended = 0x15
    Discover_Attribute_Extended_rsp = 0x16


GENERAL_COMMANDS = COMMANDS = {
    GeneralCommand.Read_Attributes: ZCLCommandDef(
        schema={"attribute_ids": t.List[t.uint16_t]},
        direction=Direction.Client_to_Server,
    ),
    GeneralCommand.Read_Attributes_rsp: ZCLCommandDef(
        schema=ReadAttributesResponse,
        direction=Direction.Server_to_Client,
    ),
    GeneralCommand.Write_Attributes: ZCLCommandDef(
        schema={"attributes": t.List[Attribute]}, direction=Direction.Client_to_Server
    ),
    GeneralCommand.Write_Attributes_Undivided: ZCLCommandDef(
        schema={"attributes": t.List[Attribute]}, direction=Direction.Client_to_Server
    ),
    GeneralCommand.Write_Attributes_rsp: ZCLCommandDef(
        schema=WriteAttributesResponseSchema,
        direction=Direction.Server_to_Client,
    ),
    GeneralCommand.Write_Attributes_No_Response: ZCLCommandDef(
        schema={"attributes": t.List[Attribute]}, direction=Direction.Client_to_Server
    ),
    GeneralCommand.Configure_Reporting: ZCLCommandDef(
        schema={"config_records": t.List[AttributeReportingConfig]},
        direction=Direction.Client_to_Server,
    ),
    GeneralCommand.Configure_Reporting_rsp: ZCLCommandDef(
        schema={"status_records": ConfigureReportingResponse},
        direction=Direction.Server_to_Client,
    ),
    GeneralCommand.Read_Reporting_Configuration: ZCLCommandDef(
        schema={"attribute_records": t.List[ReadReportingConfigRecord]},
        direction=Direction.Client_to_Server,
    ),
    GeneralCommand.Read_Reporting_Configuration_rsp: ZCLCommandDef(
        schema={"attribute_configs": t.List[AttributeReportingConfigWithStatus]},
        direction=Direction.Server_to_Client,
    ),
    GeneralCommand.Report_Attributes: ZCLCommandDef(
        schema={"attribute_reports": t.List[Attribute]},
        direction=Direction.Client_to_Server,
    ),
    GeneralCommand.Default_Response: ZCLCommandDef(
        schema=DefaultResponse,
        direction=Direction.Server_to_Client,
    ),
    GeneralCommand.Discover_Attributes: ZCLCommandDef(
        schema={"start_attribute_id": t.uint16_t, "max_attribute_ids": t.uint8_t},
        direction=Direction.Client_to_Server,
    ),
    GeneralCommand.Discover_Attributes_rsp: ZCLCommandDef(
        schema={
            "discovery_complete": t.Bool,
            "attribute_info": t.List[DiscoverAttributesResponseRecord],
        },
        direction=Direction.Server_to_Client,
    ),
    GeneralCommand.Read_Attributes_Structured: ZCLCommandDef(
        schema={"attributes": t.List[ReadAttributeStructured]},
        direction=Direction.Client_to_Server,
    ),
    GeneralCommand.Write_Attributes_Structured: ZCLCommandDef(
        schema={"attributes": t.List[WriteAttributeStructured]},
        direction=Direction.Client_to_Server,
    ),
    GeneralCommand.Write_Attributes_Structured_rsp: ZCLCommandDef(
        schema=WriteAttributesStructuredResponseSchema,
        direction=Direction.Server_to_Client,
    ),
    GeneralCommand.Discover_Commands_Received: ZCLCommandDef(
        schema={"start_command_id": t.uint8_t, "max_command_ids": t.uint8_t},
        direction=Direction.Client_to_Server,
    ),
    GeneralCommand.Discover_Commands_Received_rsp: ZCLCommandDef(
        schema={"discovery_complete": t.Bool, "command_ids": t.List[t.uint8_t]},
        direction=Direction.Server_to_Client,
    ),
    GeneralCommand.Discover_Commands_Generated: ZCLCommandDef(
        schema={"start_command_id": t.uint8_t, "max_command_ids": t.uint8_t},
        direction=Direction.Client_to_Server,
    ),
    GeneralCommand.Discover_Commands_Generated_rsp: ZCLCommandDef(
        schema={"discovery_complete": t.Bool, "command_ids": t.List[t.uint8_t]},
        direction=Direction.Server_to_Client,
    ),
    GeneralCommand.Discover_Attribute_Extended: ZCLCommandDef(
        schema={"start_attribute_id": t.uint16_t, "max_attribute_ids": t.uint8_t},
        direction=Direction.Client_to_Server,
    ),
    GeneralCommand.Discover_Attribute_Extended_rsp: ZCLCommandDef(
        schema={
            "discovery_complete": t.Bool,
            "extended_attr_info": t.List[DiscoverAttributesExtendedResponseRecord],
        },
        direction=Direction.Server_to_Client,
    ),
}

for command_id, command_def in list(GENERAL_COMMANDS.items()):
    GENERAL_COMMANDS[command_id] = command_def.replace(
        id=command_id, name=command_id.name
    ).with_compiled_schema()

ZCL_CLUSTER_REVISION_ATTR = ZCLAttributeDef(
    id=0xFFFD, type=t.uint16_t, access="r", mandatory=True
)
ZCL_REPORTING_STATUS_ATTR = ZCLAttributeDef(
    id=0xFFFE, type=AttributeReportingStatus, access="r"
)



from __future__ import annotations

import asyncio
import contextlib
from unittest import mock

import pytest

from tests.conftest import App, make_ieee, make_neighbor, make_route
import zigpy.config as conf
import zigpy.device
import zigpy.endpoint
import zigpy.profiles
import zigpy.topology
import zigpy.types as t
import zigpy.zdo.types as zdo_t


@pytest.fixture(autouse=True)
def remove_request_delay():
    with mock.patch("zigpy.topology.REQUEST_DELAY", new=(0, 0)):
        yield


@pytest.fixture
def topology(make_initialized_device):
    app = App(
        {
            conf.CONF_DEVICE: {conf.CONF_DEVICE_PATH: "/dev/null"},
            conf.CONF_TOPO_SKIP_COORDINATOR: True,
        }
    )

    coordinator = make_initialized_device(app)
    coordinator.nwk = 0x0000

    app.state.node_info.nwk = coordinator.nwk
    app.state.node_info.ieee = coordinator.ieee
    app.state.node_info.logical_type = zdo_t.LogicalType.Coordinator

    return zigpy.topology.Topology(app)


@contextlib.contextmanager
def patch_device_tables(
    device: zigpy.device.Device,
    neighbors: list | BaseException | zdo_t.Status,
    routes: list | BaseException | zdo_t.Status,
):
    def mgmt_lqi_req(StartIndex: t.uint8_t, **kwargs):
        status = zdo_t.Status.SUCCESS
        entries = 0
        start_index = 0
        table: list[zdo_t.Neighbor] = []

        if isinstance(neighbors, zdo_t.Status):
            status = neighbors
        elif isinstance(neighbors, BaseException):
            raise neighbors
        else:
            entries = len(neighbors)
            start_index = StartIndex
            table = neighbors[StartIndex : StartIndex + 3]

        return list(
            {
                "Status": status,
                "Neighbors": zdo_t.Neighbors(
                    Entries=entries,
                    StartIndex=start_index,
                    NeighborTableList=table,
                ),
            }.values()
        )

    def mgmt_rtg_req(StartIndex: t.uint8_t, **kwargs):
        status = zdo_t.Status.SUCCESS
        entries = 0
        start_index = 0
        table: list[zdo_t.Route] = []

        if isinstance(routes, zdo_t.Status):
            status = routes
        elif isinstance(routes, BaseException):
            raise routes
        else:
            entries = len(routes)
            start_index = StartIndex
            table = routes[StartIndex : StartIndex + 3]

        return list(
            {
                "Status": status,
                "Routes": zdo_t.Routes(
                    Entries=entries,
                    StartIndex=start_index,
                    RoutingTableList=table,
                ),
            }.values()
        )

    lqi_req_patch = mock.patch.object(
        device.zdo,
        "Mgmt_Lqi_req",
        mock.AsyncMock(side_effect=mgmt_lqi_req, spec_set=device.zdo.Mgmt_Lqi_req),
    )
    rtg_req_patch = mock.patch.object(
        device.zdo,
        "Mgmt_Rtg_req",
        mock.AsyncMock(side_effect=mgmt_rtg_req, spec_set=device.zdo.Mgmt_Rtg_req),
    )

    with lqi_req_patch, rtg_req_patch:
        yield


async def test_scan_no_devices(topology) -> None:
    await topology.scan()

    assert not topology.neighbors
    assert not topology.routes


@pytest.mark.parametrize(
    ("neighbors", "routes"),
    [
        ([], TimeoutError()),
        ([], []),
        (TimeoutError(), TimeoutError()),
    ],
)
async def test_scan_failures(
    topology, make_initialized_device, neighbors, routes
) -> None:
    dev = make_initialized_device(topology._app)

    with patch_device_tables(dev, neighbors=neighbors, routes=routes):
        await topology.scan()

        assert len(dev.zdo.Mgmt_Lqi_req.mock_calls) == 1 if not neighbors else 3
        assert len(dev.zdo.Mgmt_Rtg_req.mock_calls) == 1 if not routes else 3

    assert not topology.neighbors[dev.ieee]
    assert not topology.routes[dev.ieee]


async def test_neighbors_not_supported(topology, make_initialized_device) -> None:
    dev = make_initialized_device(topology._app)

    with patch_device_tables(dev, neighbors=zdo_t.Status.NOT_SUPPORTED, routes=[]):
        await topology.scan()

        assert len(dev.zdo.Mgmt_Lqi_req.mock_calls) == 1
        assert len(dev.zdo.Mgmt_Rtg_req.mock_calls) == 1

        await topology.scan()

        assert len(dev.zdo.Mgmt_Lqi_req.mock_calls) == 1
        assert len(dev.zdo.Mgmt_Rtg_req.mock_calls) == 2


async def test_routes_not_supported(topology, make_initialized_device) -> None:
    dev = make_initialized_device(topology._app)

    with patch_device_tables(dev, neighbors=[], routes=zdo_t.Status.NOT_SUPPORTED):
        await topology.scan()

        assert len(dev.zdo.Mgmt_Lqi_req.mock_calls) == 1
        assert len(dev.zdo.Mgmt_Rtg_req.mock_calls) == 1

        await topology.scan()

        assert len(dev.zdo.Mgmt_Lqi_req.mock_calls) == 2
        assert len(dev.zdo.Mgmt_Rtg_req.mock_calls) == 1


async def test_routes_and_neighbors_not_supported(
    topology, make_initialized_device
) -> None:
    dev = make_initialized_device(topology._app)

    with patch_device_tables(
        dev, neighbors=zdo_t.Status.NOT_SUPPORTED, routes=zdo_t.Status.NOT_SUPPORTED
    ):
        await topology.scan()

        assert len(dev.zdo.Mgmt_Lqi_req.mock_calls) == 1
        assert len(dev.zdo.Mgmt_Rtg_req.mock_calls) == 1

        await topology.scan()

        assert len(dev.zdo.Mgmt_Lqi_req.mock_calls) == 1
        assert len(dev.zdo.Mgmt_Rtg_req.mock_calls) == 1


async def test_scan_end_device(topology, make_initialized_device) -> None:
    dev = make_initialized_device(topology._app)
    dev.node_desc.logical_type = zdo_t.LogicalType.EndDevice

    with patch_device_tables(dev, neighbors=[], routes=[]):
        await topology.scan()

        # The device will not be scanned because it is not a router
        assert len(dev.zdo.Mgmt_Lqi_req.mock_calls) == 0
        assert len(dev.zdo.Mgmt_Rtg_req.mock_calls) == 0


async def test_scan_explicit_device(topology, make_initialized_device) -> None:
    dev1 = make_initialized_device(topology._app)
    dev2 = make_initialized_device(topology._app)

    with patch_device_tables(dev1, neighbors=[], routes=[]):
        with patch_device_tables(dev2, neighbors=[], routes=[]):
            await topology.scan(devices=[dev2])

            # Only the second device was scanned
            assert len(dev1.zdo.Mgmt_Lqi_req.mock_calls) == 0
            assert len(dev1.zdo.Mgmt_Rtg_req.mock_calls) == 0
            assert len(dev2.zdo.Mgmt_Lqi_req.mock_calls) == 1
            assert len(dev2.zdo.Mgmt_Rtg_req.mock_calls) == 1


async def test_scan_router_many(topology, make_initialized_device) -> None:
    dev = make_initialized_device(topology._app)

    with patch_device_tables(
        dev,
        neighbors=[
            make_neighbor(ieee=make_ieee(2 + i), nwk=0x1234 + i) for i in range(100)
        ],
        routes=[
            make_route(dest_nwk=0x1234 + i, next_hop=0x1234 + i) for i in range(100)
        ],
    ):
        await topology.scan()

        # We only permit three scans per request
        assert len(dev.zdo.Mgmt_Lqi_req.mock_calls) == 34
        assert len(dev.zdo.Mgmt_Rtg_req.mock_calls) == 34

    assert topology.neighbors[dev.ieee] == [
        make_neighbor(ieee=make_ieee(2 + i), nwk=0x1234 + i) for i in range(100)
    ]
    assert topology.routes[dev.ieee] == [
        make_route(dest_nwk=0x1234 + i, next_hop=0x1234 + i) for i in range(100)
    ]


async def test_scan_skip_coordinator(topology, make_initialized_device) -> None:
    coordinator = topology._app._device
    assert coordinator.nwk == 0x0000

    with patch_device_tables(coordinator, neighbors=[], routes=[]):
        await topology.scan()

        assert len(coordinator.zdo.Mgmt_Lqi_req.mock_calls) == 0
        assert len(coordinator.zdo.Mgmt_Rtg_req.mock_calls) == 0

    assert not topology.neighbors[coordinator.ieee]
    assert not topology.routes[coordinator.ieee]


async def test_scan_coordinator(topology) -> None:
    app = topology._app
    app.config[conf.CONF_TOPO_SKIP_COORDINATOR] = False

    coordinator = app._device
    coordinator.node_desc.logical_type = zdo_t.LogicalType.Coordinator
    assert coordinator.nwk == 0x0000

    with patch_device_tables(
        coordinator,
        neighbors=[
            make_neighbor(ieee=make_ieee(2), nwk=0x1234),
        ],
        routes=[
            make_route(dest_nwk=0x1234, next_hop=0x1234),
        ],
    ):
        await topology.scan()

        assert len(coordinator.zdo.Mgmt_Lqi_req.mock_calls) == 1
        assert len(coordinator.zdo.Mgmt_Rtg_req.mock_calls) == 1

    assert topology.neighbors[coordinator.ieee] == [
        make_neighbor(ieee=make_ieee(2), nwk=0x1234)
    ]
    assert topology.routes[coordinator.ieee] == [
        make_route(dest_nwk=0x1234, next_hop=0x1234)
    ]


@mock.patch("zigpy.application.ControllerApplication._discover_unknown_device")
async def test_discover_new_devices(
    discover_unknown_device, topology, make_initialized_device
) -> None:
    dev1 = make_initialized_device(topology._app)
    dev2 = make_initialized_device(topology._app)

    await topology._find_unknown_devices(
        neighbors={
            dev1.ieee: [
                # Existing devices
                make_neighbor(ieee=dev1.ieee, nwk=dev1.nwk),
                make_neighbor(ieee=dev2.ieee, nwk=dev2.nwk),
                # Unknown device
                make_neighbor(
                    ieee=t.EUI64.convert("aa:bb:cc:dd:11:22:33:44"), nwk=0xFF00
                ),
            ],
            dev2.ieee: [],
        },
        routes={
            dev1.ieee: [
                # Existing devices
                make_route(dest_nwk=dev1.nwk, next_hop=dev1.nwk),
                make_route(dest_nwk=dev2.nwk, next_hop=dev2.nwk),
                # Via existing devices
                make_route(dest_nwk=0xFF01, next_hop=dev2.nwk),
                make_route(dest_nwk=dev2.nwk, next_hop=0xFF02),
                # Inactive route
                make_route(
                    dest_nwk=0xFF03, next_hop=0xFF04, status=zdo_t.RouteStatus.Inactive
                ),
            ],
            dev2.ieee: [],
        },
    )

    assert len(discover_unknown_device.mock_calls) == 3
    assert mock.call(0xFF00) in discover_unknown_device.mock_calls
    assert mock.call(0xFF01) in discover_unknown_device.mock_calls
    assert mock.call(0xFF02) in discover_unknown_device.mock_calls


@mock.patch("zigpy.topology.Topology._scan")
async def test_scan_start_concurrent(mock_scan, topology):
    concurrency = 0
    max_concurrency = 0

    async def _scan(_):
        nonlocal concurrency
        nonlocal max_concurrency

        concurrency += 1
        max_concurrency = max(concurrency, max_concurrency)

        try:
            await asyncio.sleep(0.01)
        finally:
            concurrency -= 1
            max_concurrency = max(concurrency, max_concurrency)

    mock_scan.side_effect = _scan

    topology.start_periodic_scans(0.1)
    topology.start_periodic_scans(0.1)
    topology.start_periodic_scans(0.1)
    topology.start_periodic_scans(0.1)
    topology.start_periodic_scans(0.1)

    scan1 = asyncio.create_task(topology.scan())
    scan2 = asyncio.create_task(topology.scan())

    await asyncio.sleep(0.01)

    with pytest.raises(asyncio.CancelledError):
        await scan1

    await scan2

    # Wait for a "scan" to finish
    await asyncio.sleep(0.15)
    await topology._scan_task
    topology.stop_periodic_scans()

    # Only a single one was actually running
    assert max_concurrency == 1

    topology.stop_periodic_scans()

    await asyncio.sleep(0)

    # All of the tasks have been stopped
    assert topology._scan_task.done()
    assert topology._scan_loop_task.done()


@mock.patch("zigpy.topology.Topology.scan", side_effect=RuntimeError())
async def test_periodic_scan_failure(mock_scan, topology):
    topology.start_periodic_scans(0.01)
    await asyncio.sleep(0.1)
    topology.stop_periodic_scans()


async def test_periodic_scan_priority(topology):
    async def _scan(_):
        await asyncio.sleep(0.5)

    with mock.patch.object(topology, "_scan", side_effect=_scan) as mock_scan:
        scan_task = asyncio.create_task(topology.scan())
        await asyncio.sleep(0.1)

        # Start a periodic scan. It won't have time to run yet, the old scan is running
        topology.start_periodic_scans(0.05)

        # Wait for the original scan to finish
        await scan_task

        # Start another scan, interrupting the periodic scan
        await asyncio.sleep(0.15)
        await topology.scan()

        # Now we can cancel the periodic scan
        topology.stop_periodic_scans()
        await asyncio.sleep(0)

    # Our two manual scans succeeded and the periodic one was attempted
    assert len(mock_scan.mock_calls) == 3



from __future__ import annotations

from typing import Final

import zigpy.types as t
from zigpy.zcl import Cluster
from zigpy.zcl.foundation import (
    BaseAttributeDefs,
    BaseCommandDefs,
    DataTypeId,
    ZCLAttributeDef,
    ZCLCommandDef,
)


class Price(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0700
    ep_attribute: Final = "smartenergy_price"


class Drlc(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0701
    ep_attribute: Final = "smartenergy_drlc"


class RegisteredTier(t.enum8):
    No_Tier = 0x00
    Tier_1 = 0x01
    Tier_2 = 0x02
    Tier_3 = 0x03
    Tier_4 = 0x04
    Tier_5 = 0x05
    Tier_6 = 0x06
    Tier_7 = 0x07
    Tier_8 = 0x08
    Tier_9 = 0x09
    Tier_10 = 0x0A
    Tier_11 = 0x0B
    Tier_12 = 0x0C
    Tier_13 = 0x0D
    Tier_14 = 0x0E
    Extended_Tier = 0x0F


class MeteringDeviceType(t.enum8):
    """Metering device type."""

    Electric_Metering = 0
    Gas_Metering = 1
    Water_Metering = 2
    Thermal_Metering = 3  # Deprecated
    Pressure_Metering = 4
    Heat_Metering = 5
    Cooling_Metering = 6
    EUMD_for_metering_electric_vehicle_charging = 7
    PV_Generation_Metering = 8
    Wind_Turbine_Generation_Metering = 9
    Water_Turbine_Generation_Metering = 10
    Micro_Generation_Metering = 11
    Solar_Hot_Water_Generation_Metering = 12
    Electric_Metering_Element_Phase_1 = 13
    Electric_Metering_Element_Phase_2 = 14
    Electric_Metering_Element_Phase_3 = 15

    # 127 + above enum values
    Mirrored_Electric_Metering = 127
    Mirrored_Gas_Metering = 128
    Mirrored_Water_Metering = 129
    Mirrored_Thermal_Metering = 130  # Deprecated
    Mirrored_Pressure_Metering = 131
    Mirrored_Heat_Metering = 132
    Mirrored_Cooling_Metering = 133
    Mirrored_EUMD_for_metering_electric_vehicle_charging = 134
    Mirrored_PV_Generation_Metering = 135
    Mirrored_Wind_Turbine_Generation_Metering = 136
    Mirrored_Water_Turbine_Generation_Metering = 137
    Mirrored_Micro_Generation_Metering = 138
    Mirrored_Solar_Hot_Water_Generation_Metering = 139
    Mirrored_Electric_Metering_Element_Phase_1 = 140
    Mirrored_Electric_Metering_Element_Phase_2 = 141
    Mirrored_Electric_Metering_Element_Phase_3 = 142


class MeteringUnitofMeasure(t.enum8):
    """Metering unit of measure."""

    Kwh_and_Kwh_binary = 0x00
    Cubic_Meter_and_Cubic_Meter_per_Hour_binary = 0x01
    Cubic_Feet_and_Cubic_Feet_per_Hour_binary = 0x02
    Ccf_and_Ccf_per_Hour_binary = 0x03
    US_Gallons_and_US_Gallons_per_Hour_binary = 0x04
    Imperial_Gallons_and_Imperial_Gallons_per_Hour_binary = 0x05
    BTU_and_BTU_per_Hour_binary = 0x06
    Liters_and_Liters_per_Hour_binary = 0x07
    KPA_gauge_binary = 0x08
    KPA_absolute_binary = 0x09
    MCF_and_MCF_per_Hour_binary = 0x0A
    Unitless_binary = 0x0B
    Mega_Joule_and_Mega_Joule_per_second_binary = 0x0C
    Kvar_and_Kvarh_binary = 0x0D
    Kwh_and_Kwh_bcd = 0x80
    Cubic_Meter_and_Cubic_Meter_per_Hour_bcd = 0x81
    Cubic_Feet_and_Cubic_Feet_per_Hour_bcd = 0x82
    Ccf_and_Ccf_per_Hour_bcd = 0x83
    US_Gallons_and_US_Gallons_per_Hour_bcd = 0x84
    Imperial_Gallons_and_Imperial_Gallons_per_Hour_bcd = 0x85
    BTU_and_BTU_per_Hour_bcd = 0x86
    Liters_and_Liters_per_Hour_bcd = 0x87
    KPA_gauge_bcd = 0x88
    KPA_absolute_bcd = 0x89
    MCF_and_MCF_per_Hour_bcd = 0x8A
    Unitless_bcd = 0x8B
    Mega_Joule_and_Mega_Joule_per_second_bcd = 0x8C
    Kvar_and_Kvarh_bcd = 0x8D


class NumberFormatting(t.IntStruct, t.uint8_t):
    """Number formatting."""

    num_digits_right_of_decimal: t.uint3_t
    num_digits_left_of_decimal: t.uint4_t
    suppress_leading_zeros: t.uint1_t


class MeteringStatus(t.bitmap8):
    """Metering status."""

    Check_Meter = 0b00000001
    Low_Battery = 0b00000010
    Tamper_Detect = 0b00000100
    Power_Failure = 0b00001000
    Power_Quality = 0b00010000
    Leak_Detect = 0b00100000
    Service_Disconnect_Open = 0b01000000
    Reserved = 0b10000000


class Metering(Cluster):
    RegisteredTier: Final = RegisteredTier
    MeteringDeviceType: Final = MeteringDeviceType
    MeteringUnitofMeasure: Final = MeteringUnitofMeasure
    NumberFormatting: Final = NumberFormatting

    cluster_id: Final[t.uint16_t] = 0x0702
    ep_attribute: Final = "smartenergy_metering"

    class AttributeDefs(BaseAttributeDefs):
        current_summ_delivered: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint48_t, access="r"
        )
        current_summ_received: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint48_t, access="r"
        )
        current_max_demand_delivered: Final = ZCLAttributeDef(
            id=0x0002, type=t.uint48_t, access="r"
        )
        current_max_demand_received: Final = ZCLAttributeDef(
            id=0x0003, type=t.uint48_t, access="r"
        )
        dft_summ: Final = ZCLAttributeDef(id=0x0004, type=t.uint48_t, access="r")
        daily_freeze_time: Final = ZCLAttributeDef(
            id=0x0005, type=t.uint16_t, access="r"
        )
        power_factor: Final = ZCLAttributeDef(id=0x0006, type=t.int8s, access="r")
        reading_snapshot_time: Final = ZCLAttributeDef(
            id=0x0007, type=t.UTCTime, access="r"
        )
        current_max_demand_delivered_time: Final = ZCLAttributeDef(
            id=0x0008, type=t.UTCTime, access="r"
        )
        current_max_demand_received_time: Final = ZCLAttributeDef(
            id=0x0009, type=t.UTCTime, access="r"
        )
        default_update_period: Final = ZCLAttributeDef(
            id=0x000A, type=t.uint8_t, access="r"
        )
        fast_poll_update_period: Final = ZCLAttributeDef(
            id=0x000B, type=t.uint8_t, access="r"
        )
        current_block_period_consumption_delivered: Final = ZCLAttributeDef(
            id=0x000C, type=t.uint48_t, access="r"
        )
        daily_consumption_target: Final = ZCLAttributeDef(
            id=0x000D, type=t.uint24_t, access="r"
        )
        current_block: Final = ZCLAttributeDef(id=0x000E, type=t.enum8, access="r")
        profile_interval_period: Final = ZCLAttributeDef(
            id=0x000F, type=t.enum8, access="r"
        )
        # 0x0010: ('interval_read_reporting_period', UNKNOWN), # Deprecated
        preset_reading_time: Final = ZCLAttributeDef(
            id=0x0011, type=t.uint16_t, access="r"
        )
        volume_per_report: Final = ZCLAttributeDef(
            id=0x0012, type=t.uint16_t, access="r"
        )
        flow_restriction: Final = ZCLAttributeDef(id=0x0013, type=t.uint8_t, access="r")
        supply_status: Final = ZCLAttributeDef(id=0x0014, type=t.enum8, access="r")
        current_in_energy_carrier_summ: Final = ZCLAttributeDef(
            id=0x0015, type=t.uint48_t, access="r"
        )
        current_out_energy_carrier_summ: Final = ZCLAttributeDef(
            id=0x0016, type=t.uint48_t, access="r"
        )
        inlet_temperature: Final = ZCLAttributeDef(id=0x0017, type=t.int24s, access="r")
        outlet_temperature: Final = ZCLAttributeDef(
            id=0x0018, type=t.int24s, access="r"
        )
        control_temperature: Final = ZCLAttributeDef(
            id=0x0019, type=t.int24s, access="r"
        )
        current_in_energy_carrier_demand: Final = ZCLAttributeDef(
            id=0x001A, type=t.int24s, access="r"
        )
        current_out_energy_carrier_demand: Final = ZCLAttributeDef(
            id=0x001B, type=t.int24s, access="r"
        )
        current_block_period_consumption_received: Final = ZCLAttributeDef(
            id=0x001D, type=t.uint48_t, access="r"
        )
        current_block_received: Final = ZCLAttributeDef(
            id=0x001E, type=t.uint48_t, access="r"
        )
        dft_summation_received: Final = ZCLAttributeDef(
            id=0x001F, type=t.uint48_t, access="r"
        )
        active_register_tier_delivered: Final = ZCLAttributeDef(
            id=0x0020, type=RegisteredTier, access="r"
        )
        active_register_tier_received: Final = ZCLAttributeDef(
            id=0x0021, type=RegisteredTier, access="r"
        )
        last_block_switch_time: Final = ZCLAttributeDef(
            id=0x0022, type=t.UTCTime, access="r"
        )
        # 0x0100: ('change_reporting_profile', UNKNOWN),
        current_tier1_summ_delivered: Final = ZCLAttributeDef(
            id=0x0100, type=t.uint48_t, access="r"
        )
        current_tier1_summ_received: Final = ZCLAttributeDef(
            id=0x0101, type=t.uint48_t, access="r"
        )
        current_tier2_summ_delivered: Final = ZCLAttributeDef(
            id=0x0102, type=t.uint48_t, access="r"
        )
        current_tier2_summ_received: Final = ZCLAttributeDef(
            id=0x0103, type=t.uint48_t, access="r"
        )
        current_tier3_summ_delivered: Final = ZCLAttributeDef(
            id=0x0104, type=t.uint48_t, access="r"
        )
        current_tier3_summ_received: Final = ZCLAttributeDef(
            id=0x0105, type=t.uint48_t, access="r"
        )
        current_tier4_summ_delivered: Final = ZCLAttributeDef(
            id=0x0106, type=t.uint48_t, access="r"
        )
        current_tier4_summ_received: Final = ZCLAttributeDef(
            id=0x0107, type=t.uint48_t, access="r"
        )
        current_tier5_summ_delivered: Final = ZCLAttributeDef(
            id=0x0108, type=t.uint48_t, access="r"
        )
        current_tier5_summ_received: Final = ZCLAttributeDef(
            id=0x0109, type=t.uint48_t, access="r"
        )
        current_tier6_summ_delivered: Final = ZCLAttributeDef(
            id=0x010A, type=t.uint48_t, access="r"
        )
        current_tier6_summ_received: Final = ZCLAttributeDef(
            id=0x010B, type=t.uint48_t, access="r"
        )
        current_tier7_summ_delivered: Final = ZCLAttributeDef(
            id=0x010C, type=t.uint48_t, access="r"
        )
        current_tier7_summ_received: Final = ZCLAttributeDef(
            id=0x010D, type=t.uint48_t, access="r"
        )
        current_tier8_summ_delivered: Final = ZCLAttributeDef(
            id=0x010E, type=t.uint48_t, access="r"
        )
        current_tier8_summ_received: Final = ZCLAttributeDef(
            id=0x010F, type=t.uint48_t, access="r"
        )
        current_tier9_summ_delivered: Final = ZCLAttributeDef(
            id=0x0110, type=t.uint48_t, access="r"
        )
        current_tier9_summ_received: Final = ZCLAttributeDef(
            id=0x0111, type=t.uint48_t, access="r"
        )
        current_tier10_summ_delivered: Final = ZCLAttributeDef(
            id=0x0112, type=t.uint48_t, access="r"
        )
        current_tier10_summ_received: Final = ZCLAttributeDef(
            id=0x0113, type=t.uint48_t, access="r"
        )
        current_tier11_summ_delivered: Final = ZCLAttributeDef(
            id=0x0114, type=t.uint48_t, access="r"
        )
        current_tier11_summ_received: Final = ZCLAttributeDef(
            id=0x0115, type=t.uint48_t, access="r"
        )
        current_tier12_summ_delivered: Final = ZCLAttributeDef(
            id=0x0116, type=t.uint48_t, access="r"
        )
        current_tier12_summ_received: Final = ZCLAttributeDef(
            id=0x0117, type=t.uint48_t, access="r"
        )
        current_tier13_summ_delivered: Final = ZCLAttributeDef(
            id=0x0118, type=t.uint48_t, access="r"
        )
        current_tier13_summ_received: Final = ZCLAttributeDef(
            id=0x0119, type=t.uint48_t, access="r"
        )
        current_tier14_summ_delivered: Final = ZCLAttributeDef(
            id=0x011A, type=t.uint48_t, access="r"
        )
        current_tier14_summ_received: Final = ZCLAttributeDef(
            id=0x011B, type=t.uint48_t, access="r"
        )
        current_tier15_summ_delivered: Final = ZCLAttributeDef(
            id=0x011C, type=t.uint48_t, access="r"
        )
        current_tier15_summ_received: Final = ZCLAttributeDef(
            id=0x011D, type=t.uint48_t, access="r"
        )
        status: Final = ZCLAttributeDef(id=0x0200, type=MeteringStatus, access="r")
        remaining_battery_life: Final = ZCLAttributeDef(
            id=0x0201, type=t.uint8_t, access="r"
        )
        hours_in_operation: Final = ZCLAttributeDef(
            id=0x0202, type=t.uint24_t, access="r"
        )
        hours_in_fault: Final = ZCLAttributeDef(id=0x0203, type=t.uint24_t, access="r")
        extended_status: Final = ZCLAttributeDef(id=0x0204, type=t.bitmap64, access="r")
        remaining_battery_life_days: Final = ZCLAttributeDef(
            id=0x0205, type=t.uint16_t, access="r"
        )
        current_meter_id: Final = ZCLAttributeDef(id=0x0206, type=t.LVBytes, access="r")
        iambient_consumption_indicator: Final = ZCLAttributeDef(
            id=0x0207, type=t.enum8, access="r"
        )
        unit_of_measure: Final = ZCLAttributeDef(
            id=0x0300, type=MeteringUnitofMeasure, access="r"
        )
        multiplier: Final = ZCLAttributeDef(id=0x0301, type=t.uint24_t, access="r")
        divisor: Final = ZCLAttributeDef(id=0x0302, type=t.uint24_t, access="r")

        # This attribute shall be used against the following attributes:
        # • CurrentSummationDelivered
        # • CurrentSummationReceived
        # • SummationDeliveredPerReport
        # • TOU Information attributes
        # • DFTSummation
        # • Block Information attributes
        summation_formatting: Final = ZCLAttributeDef(
            id=0x0303, zcl_type=DataTypeId.map8, type=NumberFormatting, access="r"
        )

        # This attribute shall be used against the following attributes:
        # • CurrentMaxDemandDelivered
        # • CurrentMaxDemandReceived
        # • InstantaneousDemand
        demand_formatting: Final = ZCLAttributeDef(
            id=0x0304, zcl_type=DataTypeId.map8, type=NumberFormatting, access="r"
        )

        # This attribute shall be used against the following attributes:
        # • CurrentDayConsumptionDelivered
        # • CurrentDayConsumptionReceived
        # • PreviousDayConsumptionDelivered
        # • PreviousDayConsumptionReceived
        # • CurrentPartialProfileIntervalValue
        # • Intervals
        # • DailyConsumptionTarget
        # • CurrentDayConsumptionDelivered
        # • CurrentDayConsumptionReceived
        # • PreviousDayNConsumptionDelivered
        # • PreviousDayNConsumptionReceived
        # • CurrentWeekConsumptionDelivered
        # • CurrentWeekConsumptionReceived
        # • PreviousWeekNConsumptionDelivered
        # • PreviousWeekNConsumptionReceived
        # • CurrentMonthConsumptionDelivered
        # • CurrentMonthConsumptionReceived
        # • PreviousMonthNConsumptionDelivered
        # • PreviousMonthNConsumptionReceived
        historical_consumption_formatting: Final = ZCLAttributeDef(
            id=0x0305, zcl_type=DataTypeId.map8, type=NumberFormatting, access="r"
        )
        metering_device_type: Final = ZCLAttributeDef(
            id=0x0306,
            type=MeteringDeviceType,
            # Note that these values represent an Enumeration, and not an 8-bit bitmap
            # as indicated in the attribute description. For backwards compatibility
            # reasons, the data type has not been changed, though the data itself should
            # be treated like an enum
            zcl_type=DataTypeId.map8,
            access="r",
        )
        site_id: Final = ZCLAttributeDef(
            id=0x0307, type=t.LimitedLVBytes(32), access="r"
        )
        meter_serial_number: Final = ZCLAttributeDef(
            id=0x0308, type=t.LimitedLVBytes(24), access="r"
        )
        energy_carrier_unit_of_measure: Final = ZCLAttributeDef(
            id=0x0309, type=MeteringUnitofMeasure, access="r"
        )
        energy_carrier_summation_formatting: Final = ZCLAttributeDef(
            id=0x030A, zcl_type=DataTypeId.map8, type=NumberFormatting, access="r"
        )
        energy_carrier_demand_formatting: Final = ZCLAttributeDef(
            id=0x030B, zcl_type=DataTypeId.map8, type=NumberFormatting, access="r"
        )
        temperature_unit_of_measure: Final = ZCLAttributeDef(
            id=0x030C, type=MeteringUnitofMeasure, access="r"
        )

        # This attribute shall be used in relation with the following attributes:
        # • InletTemperature
        # • OutletTemperature
        # • ControlTemperature
        temperature_formatting: Final = ZCLAttributeDef(
            id=0x030D, zcl_type=DataTypeId.map8, type=NumberFormatting, access="r"
        )
        module_serial_number: Final = ZCLAttributeDef(
            id=0x030E, type=t.LimitedLVBytes(24), access="r"
        )
        operating_tariff_label_delivered: Final = ZCLAttributeDef(
            id=0x030F, type=t.LimitedLVBytes(24), access="r"
        )
        operating_tariff_label_received: Final = ZCLAttributeDef(
            id=0x0310, type=t.LimitedLVBytes(24), access="r"
        )
        customer_id_number: Final = ZCLAttributeDef(
            id=0x0311, type=t.LimitedLVBytes(24), access="r"
        )
        alternative_unit_of_measure: Final = ZCLAttributeDef(
            id=0x0312, type=MeteringUnitofMeasure, access="r"
        )

        # This attribute shall be used against the following attribute:
        # • AlternativeInstantaneousDemand
        alternative_demand_formatting: Final = ZCLAttributeDef(
            id=0x0313, zcl_type=DataTypeId.map8, type=NumberFormatting, access="r"
        )
        # This attribute shall be used against the following attributes:
        # • CurrentDayAlternativeConsumptionDelivered
        # • CurrentDayAlternativeConsumptionReceived
        # • PreviousDayAlternativeConsumptionDelivered
        # • PreviousDayAlternativeConsumptionReceived
        # • CurrentAlternativePartialProfileIntervalValue
        # • PreviousDayNAlternativeConsumptionDelivered
        # • PreviousDayNAlternativeConsumptionReceived
        # • CurrentWeekAlternativeConsumptionDelivered
        # • CurrentWeekAlternativeConsumptionReceived
        # • PreviousWeekNAlternativeConsumptionDelivered
        # • PreviousWeekNAlternativeConsumptionReceived
        # • CurrentMonthAlternativeConsumptionDelivered
        # • CurrentMonthAlternativeConsumptionReceived
        # • PreviousMonthNAlternativeConsumptionDelivered
        # • PreviousMonthNAlternativeConsumptionReceived
        alternative_consumption_formatting: Final = ZCLAttributeDef(
            id=0x0314, zcl_type=DataTypeId.map8, type=NumberFormatting, access="r"
        )
        instantaneous_demand: Final = ZCLAttributeDef(
            id=0x0400, type=t.int24s, access="r"
        )
        currentday_consumption_delivered: Final = ZCLAttributeDef(
            id=0x0401, type=t.uint24_t, access="r"
        )
        currentday_consumption_received: Final = ZCLAttributeDef(
            id=0x0402, type=t.uint24_t, access="r"
        )
        previousday_consumption_delivered: Final = ZCLAttributeDef(
            id=0x0403, type=t.uint24_t, access="r"
        )
        previousday_consumption_received: Final = ZCLAttributeDef(
            id=0x0404, type=t.uint24_t, access="r"
        )
        cur_part_profile_int_start_time_delivered: Final = ZCLAttributeDef(
            id=0x0405, type=t.uint32_t, access="r"
        )
        cur_part_profile_int_start_time_received: Final = ZCLAttributeDef(
            id=0x0406, type=t.uint32_t, access="r"
        )
        cur_part_profile_int_value_delivered: Final = ZCLAttributeDef(
            id=0x0407, type=t.uint24_t, access="r"
        )
        cur_part_profile_int_value_received: Final = ZCLAttributeDef(
            id=0x0408, type=t.uint24_t, access="r"
        )
        current_day_max_pressure: Final = ZCLAttributeDef(
            id=0x0409, type=t.uint48_t, access="r"
        )
        current_day_min_pressure: Final = ZCLAttributeDef(
            id=0x040A, type=t.uint48_t, access="r"
        )
        previous_day_max_pressure: Final = ZCLAttributeDef(
            id=0x040B, type=t.uint48_t, access="r"
        )
        previous_day_min_pressure: Final = ZCLAttributeDef(
            id=0x040C, type=t.uint48_t, access="r"
        )
        current_day_max_demand: Final = ZCLAttributeDef(
            id=0x040D, type=t.int24s, access="r"
        )
        previous_day_max_demand: Final = ZCLAttributeDef(
            id=0x040E, type=t.int24s, access="r"
        )
        current_month_max_demand: Final = ZCLAttributeDef(
            id=0x040F, type=t.int24s, access="r"
        )
        current_year_max_demand: Final = ZCLAttributeDef(
            id=0x0410, type=t.int24s, access="r"
        )
        currentday_max_energy_carr_demand: Final = ZCLAttributeDef(
            id=0x0411, type=t.int24s, access="r"
        )
        previousday_max_energy_carr_demand: Final = ZCLAttributeDef(
            id=0x0412, type=t.int24s, access="r"
        )
        cur_month_max_energy_carr_demand: Final = ZCLAttributeDef(
            id=0x0413, type=t.int24s, access="r"
        )
        cur_month_min_energy_carr_demand: Final = ZCLAttributeDef(
            id=0x0414, type=t.int24s, access="r"
        )
        cur_year_max_energy_carr_demand: Final = ZCLAttributeDef(
            id=0x0415, type=t.int24s, access="r"
        )
        cur_year_min_energy_carr_demand: Final = ZCLAttributeDef(
            id=0x0416, type=t.int24s, access="r"
        )
        max_number_of_periods_delivered: Final = ZCLAttributeDef(
            id=0x0500, type=t.uint8_t, access="r"
        )
        current_demand_delivered: Final = ZCLAttributeDef(
            id=0x0600, type=t.uint24_t, access="r"
        )
        demand_limit: Final = ZCLAttributeDef(id=0x0601, type=t.uint24_t, access="r")
        demand_integration_period: Final = ZCLAttributeDef(
            id=0x0602, type=t.uint8_t, access="r"
        )
        number_of_demand_subintervals: Final = ZCLAttributeDef(
            id=0x0603, type=t.uint8_t, access="r"
        )
        demand_limit_arm_duration: Final = ZCLAttributeDef(
            id=0x0604, type=t.uint16_t, access="r"
        )
        generic_alarm_mask: Final = ZCLAttributeDef(
            id=0x0800, type=t.bitmap16, access="r"
        )
        electricity_alarm_mask: Final = ZCLAttributeDef(
            id=0x0801, type=t.bitmap32, access="r"
        )
        gen_flow_pressure_alarm_mask: Final = ZCLAttributeDef(
            id=0x0802, type=t.bitmap16, access="r"
        )
        water_specific_alarm_mask: Final = ZCLAttributeDef(
            id=0x0803, type=t.bitmap16, access="r"
        )
        heat_cool_specific_alarm_mask: Final = ZCLAttributeDef(
            id=0x0804, type=t.bitmap16, access="r"
        )
        gas_specific_alarm_mask: Final = ZCLAttributeDef(
            id=0x0805, type=t.bitmap16, access="r"
        )
        extended_generic_alarm_mask: Final = ZCLAttributeDef(
            id=0x0806, type=t.bitmap48, access="r"
        )
        manufacture_alarm_mask: Final = ZCLAttributeDef(
            id=0x0807, type=t.bitmap16, access="r"
        )
        bill_to_date: Final = ZCLAttributeDef(id=0x0A00, type=t.uint32_t, access="r")
        bill_to_date_time_stamp: Final = ZCLAttributeDef(
            id=0x0A01, type=t.uint32_t, access="r"
        )
        projected_bill: Final = ZCLAttributeDef(id=0x0A02, type=t.uint32_t, access="r")
        projected_bill_time_stamp: Final = ZCLAttributeDef(
            id=0x0A03, type=t.uint32_t, access="r"
        )

    class ServerCommandDefs(BaseCommandDefs):
        get_profile: Final = ZCLCommandDef(id=0x00, schema={})
        req_mirror: Final = ZCLCommandDef(id=0x01, schema={})
        mirror_rem: Final = ZCLCommandDef(id=0x02, schema={})
        req_fast_poll_mode: Final = ZCLCommandDef(id=0x03, schema={})
        get_snapshot: Final = ZCLCommandDef(id=0x04, schema={})
        take_snapshot: Final = ZCLCommandDef(id=0x05, schema={})
        mirror_report_attr_response: Final = ZCLCommandDef(id=0x06, schema={})

    class ClientCommandDefs(BaseCommandDefs):
        get_profile_response: Final = ZCLCommandDef(id=0x00, schema={})
        req_mirror_response: Final = ZCLCommandDef(id=0x01, schema={})
        mirror_rem_response: Final = ZCLCommandDef(id=0x02, schema={})
        req_fast_poll_mode_response: Final = ZCLCommandDef(id=0x03, schema={})
        get_snapshot_response: Final = ZCLCommandDef(id=0x04, schema={})


class Messaging(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0703
    ep_attribute: Final = "smartenergy_messaging"


class Tunneling(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0704
    ep_attribute: Final = "smartenergy_tunneling"


class Prepayment(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0705
    ep_attribute: Final = "smartenergy_prepayment"


class EnergyManagement(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0706
    ep_attribute: Final = "smartenergy_energy_management"


class Calendar(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0707
    ep_attribute: Final = "smartenergy_calendar"


class DeviceManagement(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0708
    ep_attribute: Final = "smartenergy_device_management"


class Events(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0709
    ep_attribute: Final = "smartenergy_events"


class MduPairing(Cluster):
    cluster_id: Final[t.uint16_t] = 0x070A
    ep_attribute: Final = "smartenergy_mdu_pairing"


class KeyEstablishment(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0800
    ep_attribute: Final = "smartenergy_key_establishment"



"""Compatibility shims: entity constants and units moved into ZHA.

`EntityType` and `EntityPlatform` live in `zha.application`; units and other
measurement constants live in `zha.units`.
"""

from __future__ import annotations

import importlib
import typing
import warnings

_MOVED_MODULES = ("zha.application", "zha.units")


def __getattr__(name: str) -> typing.Any:
    for module_path in _MOVED_MODULES:
        module = importlib.import_module(module_path)
        if hasattr(module, name):
            warnings.warn(
                f"`{__name__}.{name}` is deprecated, import it from"
                f" `{module_path}` instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(module, name)

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""CPython semaphore unit tests minimally modified to work with RequestLimiter."""

# ruff: noqa: PT009, PT027

import asyncio
import re
import unittest

from zigpy.datastructures import RequestLimiter

STR_RGX_REPR = (
    r"^<(?P<class>.*?) object at (?P<address>.*?)"
    r"\[(?P<extras>"
    r"(set|unset|locked|unlocked|filling|draining|resetting|broken)"
    r"(, value:\d)?"
    r"(, waiters:\d+)?"
    r"(, waiters:\d+\/\d+)?"  # barrier
    r")\]>\Z"
)
RGX_REPR = re.compile(STR_RGX_REPR)


class SemaphoreTests(unittest.IsolatedAsyncioTestCase):
    def test_initial_value_zero(self):
        sem = RequestLimiter(max_concurrency=0, capacities={1: 1.0})
        self.assertTrue(sem.locked(priority=1))

    async def test_repr(self):
        sem = RequestLimiter(max_concurrency=1, capacities={1: 1.0})
        # RequestLimiter format: <RequestLimiter(max_concurrency=1, active=0, waiting=0)>
        self.assertIn("max_concurrency=1", repr(sem))
        self.assertIn("active=0", repr(sem))
        self.assertIn("waiting=0", repr(sem))

        await sem._acquire(priority=1)
        self.assertIn("active=1", repr(sem))
        self.assertIn("waiting=0", repr(sem))

        # Start tasks that will wait since semaphore is already acquired
        task1 = asyncio.create_task(sem._acquire(priority=1))
        await asyncio.sleep(0)  # Let task1 start and get queued
        self.assertIn("waiting=1", repr(sem))

        task2 = asyncio.create_task(sem._acquire(priority=1))
        await asyncio.sleep(0)  # Let task2 start and get queued
        self.assertIn("waiting=2", repr(sem))

        # Clean up
        task1.cancel()
        task2.cancel()

        await asyncio.gather(task1, task2, return_exceptions=True)

    async def test_semaphore(self):
        sem = RequestLimiter(max_concurrency=1, capacities={1: 1.0})
        # self.assertEqual(1, sem.active_requests)

        with self.assertRaisesRegex(
            TypeError,
            (
                # Used up until 3.14
                r"object RequestLimiter can't be used in 'await' expression"
                # 3.14+
                r"|'RequestLimiter' object can't be awaited"
            ),
        ):
            await sem

        self.assertFalse(sem.locked(priority=1))
        # self.assertEqual(1, sem.active_requests)

    def test_semaphore_value(self):
        self.assertRaises(ValueError, RequestLimiter, -1, {})

    async def test_acquire(self):
        sem = RequestLimiter(max_concurrency=3, capacities={1: 1.0})
        result = []

        self.assertTrue(await sem._acquire(priority=1))
        self.assertTrue(await sem._acquire(priority=1))
        self.assertFalse(sem.locked(priority=1))

        async def c1(result):
            await sem._acquire(priority=1)
            result.append(1)
            return True

        async def c2(result):
            await sem._acquire(priority=1)
            result.append(2)
            return True

        async def c3(result):
            await sem._acquire(priority=1)
            result.append(3)
            return True

        async def c4(result):
            await sem._acquire(priority=1)
            result.append(4)
            return True

        t1 = asyncio.create_task(c1(result))
        t2 = asyncio.create_task(c2(result))
        t3 = asyncio.create_task(c3(result))

        await asyncio.sleep(0)
        self.assertEqual([1], result)
        self.assertTrue(sem.locked(priority=1))
        self.assertEqual(2, sem.waiting_requests)
        # self.assertEqual(0, sem.active_requests)

        t4 = asyncio.create_task(c4(result))

        sem._release(priority=1)
        sem._release(priority=1)
        # self.assertEqual(3, sem.active_requests)

        await asyncio.sleep(0)
        # self.assertEqual(0, sem.active_requests)
        self.assertEqual(3, len(result))
        self.assertTrue(sem.locked(priority=1))
        self.assertEqual(1, sem.waiting_requests)
        # self.assertEqual(0, sem.active_requests)

        self.assertTrue(t1.done())
        self.assertTrue(t1.result())
        race_tasks = [t2, t3, t4]
        done_tasks = [t for t in race_tasks if t.done() and t.result()]
        self.assertEqual(2, len(done_tasks))

        # cleanup locked semaphore
        sem._release(priority=1)
        await asyncio.gather(*race_tasks)

    async def test_acquire_cancel(self):
        sem = RequestLimiter(max_concurrency=1, capacities={1: 1.0})
        await sem._acquire(priority=1)

        acquire = asyncio.create_task(sem._acquire(priority=1))
        asyncio.get_running_loop().call_soon(acquire.cancel)
        with self.assertRaises(asyncio.CancelledError):
            await acquire
        self.assertTrue(
            (not sem._waiters) or all(waiter.done() for waiter in sem._waiters)
        )

    async def test_acquire_cancel_before_awoken(self):
        sem = RequestLimiter(max_concurrency=1, capacities={1: 1.0})

        t1 = asyncio.create_task(sem._acquire(priority=1))
        t2 = asyncio.create_task(sem._acquire(priority=1))
        t3 = asyncio.create_task(sem._acquire(priority=1))
        t4 = asyncio.create_task(sem._acquire(priority=1))

        await asyncio.sleep(0)

        t1.cancel()
        t2.cancel()
        sem._release(priority=1)

        await asyncio.sleep(0)
        await asyncio.sleep(0)
        num_done = sum(t.done() for t in [t3, t4])
        self.assertEqual(num_done, 1)
        self.assertTrue(t3.done())
        self.assertFalse(t4.done())

        t3.cancel()
        t4.cancel()
        await asyncio.sleep(0)

    async def test_acquire_hang(self):
        sem = RequestLimiter(max_concurrency=1, capacities={1: 1.0})

        t1 = asyncio.create_task(sem._acquire(priority=1))
        t2 = asyncio.create_task(sem._acquire(priority=1))
        await asyncio.sleep(0)

        t1.cancel()
        sem._release(priority=1)
        await asyncio.sleep(0)
        await asyncio.sleep(0)
        self.assertTrue(sem.locked(priority=1))
        self.assertTrue(t2.done())

    async def test_acquire_no_hang(self):
        sem = RequestLimiter(max_concurrency=1, capacities={1: 1.0})

        async def c1():
            async with sem(priority=1):
                await asyncio.sleep(0)
            t2.cancel()

        async def c2():
            async with sem(priority=1):
                self.assertFalse(True)

        t1 = asyncio.create_task(c1())
        t2 = asyncio.create_task(c2())

        r1, r2 = await asyncio.gather(t1, t2, return_exceptions=True)
        self.assertTrue(r1 is None)
        self.assertTrue(isinstance(r2, asyncio.CancelledError))

        await asyncio.wait_for(sem._acquire(priority=1), timeout=1.0)

    def test_release_not_acquired(self):
        sem = asyncio.BoundedSemaphore()

        self.assertRaises(ValueError, sem.release)

    async def test_release_no_waiters(self):
        sem = RequestLimiter(max_concurrency=1, capacities={1: 1.0})
        await sem._acquire(priority=1)
        self.assertTrue(sem.locked(priority=1))

        sem._release(priority=1)
        self.assertFalse(sem.locked(priority=1))

    async def test_acquire_fifo_order(self):
        sem = RequestLimiter(max_concurrency=1, capacities={1: 1.0})
        result = []

        async def coro(tag):
            await sem._acquire(priority=1)
            result.append(f"{tag}_1")
            await asyncio.sleep(0.01)
            sem._release(priority=1)

            await sem._acquire(priority=1)
            result.append(f"{tag}_2")
            await asyncio.sleep(0.01)
            sem._release(priority=1)

        tasks = []
        tasks.append(asyncio.create_task(coro("c1")))
        tasks.append(asyncio.create_task(coro("c2")))
        tasks.append(asyncio.create_task(coro("c3")))
        await asyncio.gather(*tasks, return_exceptions=True)

        self.assertEqual(["c1_1", "c2_1", "c3_1", "c1_2", "c2_2", "c3_2"], result)

    async def test_acquire_fifo_order_2(self):
        sem = RequestLimiter(max_concurrency=1, capacities={1: 1.0})
        result = []

        async def c1(result):
            await sem._acquire(priority=1)
            result.append(1)
            return True

        async def c2(result):
            await sem._acquire(priority=1)
            result.append(2)
            sem._release(priority=1)
            await sem._acquire(priority=1)
            result.append(4)
            return True

        async def c3(result):
            await sem._acquire(priority=1)
            result.append(3)
            return True

        t1 = asyncio.create_task(c1(result))
        t2 = asyncio.create_task(c2(result))
        t3 = asyncio.create_task(c3(result))

        await asyncio.sleep(0)

        sem._release(priority=1)
        sem._release(priority=1)

        tasks = [t1, t2, t3]
        await asyncio.gather(*tasks)
        # self.assertEqual([1, 2, 3, 4], result)
        self.assertEqual([1, 2, 4, 3], result)  # We differ here

    async def test_acquire_fifo_order_3(self):
        sem = RequestLimiter(max_concurrency=1, capacities={1: 1.0})
        result = []

        async def c1(result):
            await sem._acquire(priority=1)
            result.append(1)
            return True

        async def c2(result):
            await sem._acquire(priority=1)
            result.append(2)
            return True

        async def c3(result):
            await sem._acquire(priority=1)
            result.append(3)
            return True

        t1 = asyncio.create_task(c1(result))
        t2 = asyncio.create_task(c2(result))
        t3 = asyncio.create_task(c3(result))

        await asyncio.sleep(0)

        t1.cancel()

        await asyncio.sleep(0)

        sem._release(priority=1)
        sem._release(priority=1)

        tasks = [t1, t2, t3]
        await asyncio.gather(*tasks, return_exceptions=True)
        # self.assertEqual([2, 3], result)
        self.assertEqual([1, 2, 3], result)  # We differ here



from __future__ import annotations

import asyncio
import pathlib
from unittest.mock import patch

import pytest
import voluptuous as vol

from tests.conftest import make_app
from zigpy import config
import zigpy.device
import zigpy.ota
import zigpy.types as t


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
async def test_ota_disabled_legacy(tmp_path: pathlib.Path) -> None:
    (tmp_path / "index.json").write_text("{}")

    # Enable all the providers
    ota = zigpy.ota.OTA(
        config=config.SCHEMA_OTA(
            {
                config.CONF_OTA_ENABLED: False,  # But disable OTA
                config.CONF_OTA_ADVANCED_DIR: tmp_path,
                config.CONF_OTA_ALLOW_ADVANCED_DIR: config.CONF_OTA_ALLOW_ADVANCED_DIR_STRING,
                config.CONF_OTA_IKEA: True,
                config.CONF_OTA_INOVELLI: True,
                config.CONF_OTA_LEDVANCE: True,
                config.CONF_OTA_SALUS: True,
                config.CONF_OTA_SONOFF: True,
                config.CONF_OTA_THIRDREALITY: True,
                config.CONF_OTA_PROVIDERS: [],
                config.CONF_OTA_EXTRA_PROVIDERS: [],
                config.CONF_OTA_REMOTE_PROVIDERS: [
                    {
                        config.CONF_OTA_PROVIDER_URL: "https://example.org/remote_index.json",
                        config.CONF_OTA_PROVIDER_MANUF_IDS: [0x1234, 4476],
                    }
                ],
                config.CONF_OTA_Z2M_LOCAL_INDEX: tmp_path / "index.json",
                config.CONF_OTA_Z2M_REMOTE_INDEX: "https://example.org/z2m_index.json",
            }
        ),
        application=None,
    )

    # None are actually enabled
    assert not ota._providers


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
async def test_ota_enabled_legacy(tmp_path: pathlib.Path) -> None:
    (tmp_path / "index.json").write_text("{}")

    # Enable all the providers
    ota = zigpy.ota.OTA(
        config=config.SCHEMA_OTA(
            {
                config.CONF_OTA_ENABLED: True,
                config.CONF_OTA_BROADCAST_ENABLED: False,
                config.CONF_OTA_ADVANCED_DIR: tmp_path,
                config.CONF_OTA_ALLOW_ADVANCED_DIR: config.CONF_OTA_ALLOW_ADVANCED_DIR_STRING,
                config.CONF_OTA_IKEA: True,
                config.CONF_OTA_INOVELLI: True,
                config.CONF_OTA_LEDVANCE: True,
                config.CONF_OTA_SALUS: True,
                config.CONF_OTA_SONOFF: True,
                config.CONF_OTA_THIRDREALITY: True,
                config.CONF_OTA_PROVIDERS: [],
                config.CONF_OTA_EXTRA_PROVIDERS: [],
                config.CONF_OTA_REMOTE_PROVIDERS: [
                    {
                        config.CONF_OTA_PROVIDER_URL: "https://example.org/remote_index.json",
                        config.CONF_OTA_PROVIDER_MANUF_IDS: [0x1234, 4476],
                    }
                ],
                config.CONF_OTA_Z2M_LOCAL_INDEX: tmp_path / "index.json",
                config.CONF_OTA_Z2M_REMOTE_INDEX: "https://example.org/z2m_index.json",
            }
        ),
        application=None,
    )

    # All are enabled
    assert len(ota._providers) == 7


async def test_ota_config(tmp_path: pathlib.Path) -> None:
    # Enable all the providers
    ota = zigpy.ota.OTA(
        config=config.SCHEMA_OTA(
            {
                config.CONF_OTA_ENABLED: True,
                config.CONF_OTA_BROADCAST_ENABLED: False,
                config.CONF_OTA_EXTRA_PROVIDERS: [
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "ikea",
                        config.CONF_OTA_PROVIDER_OVERRIDE_PREVIOUS: True,
                    }
                ],
            }
        ),
        application=None,
    )

    assert ota._providers == [
        zigpy.ota.providers.ZigpyOtaProvider(),
        zigpy.ota.providers.Ledvance(),
        zigpy.ota.providers.Sonoff(),
        zigpy.ota.providers.Tradfri(),
    ]


async def test_ota_config_invalid_message(tmp_path: pathlib.Path) -> None:
    with pytest.raises(vol.Invalid):
        zigpy.ota.OTA(
            config=config.SCHEMA_OTA(
                {
                    config.CONF_OTA_ENABLED: True,
                    config.CONF_OTA_BROADCAST_ENABLED: False,
                    config.CONF_OTA_PROVIDERS: [
                        {
                            config.CONF_OTA_PROVIDER_TYPE: "advanced",
                            config.CONF_OTA_PROVIDER_WARNING: "oops",
                            config.CONF_OTA_PROVIDER_PATH: tmp_path,
                        }
                    ],
                }
            ),
            application=None,
        )


async def test_ota_config_invalid_provider(tmp_path: pathlib.Path) -> None:
    with pytest.raises(vol.Invalid):
        zigpy.ota.OTA(
            config=config.SCHEMA_OTA(
                {
                    config.CONF_OTA_ENABLED: True,
                    config.CONF_OTA_BROADCAST_ENABLED: False,
                    config.CONF_OTA_PROVIDERS: [
                        {
                            config.CONF_OTA_PROVIDER_TYPE: "oops",
                        }
                    ],
                }
            ),
            application=None,
        )


async def test_ota_config_complex(tmp_path: pathlib.Path) -> None:
    # Enable all the providers
    (tmp_path / "index.json").write_text("{}")

    ota = zigpy.ota.OTA(
        config=config.SCHEMA_OTA(
            {
                config.CONF_OTA_ENABLED: True,
                config.CONF_OTA_BROADCAST_ENABLED: False,
                config.CONF_OTA_DISABLE_DEFAULT_PROVIDERS: [
                    "zigpy_ota",
                    "ikea",
                    "sonoff",
                    "ledvance",
                ],
                config.CONF_OTA_EXTRA_PROVIDERS: [
                    # test salus provider stub
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "salus",
                        config.CONF_OTA_PROVIDER_URL: "https://salus.example.org/",
                    },
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "ikea",
                        config.CONF_OTA_PROVIDER_URL: "https://ikea1.example.org/",
                    },
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "ikea",
                        config.CONF_OTA_PROVIDER_URL: "https://ikea2.example.org/",
                        config.CONF_OTA_PROVIDER_MANUF_IDS: [0x1234, 0x5678],
                    },
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "z2m",
                        config.CONF_OTA_PROVIDER_URL: "https://z2m.example.org/",
                    },
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "ikea",
                        config.CONF_OTA_PROVIDER_OVERRIDE_PREVIOUS: True,
                        config.CONF_OTA_PROVIDER_URL: "https://ikea3.example.org/",
                        config.CONF_OTA_PROVIDER_MANUF_IDS: [0xABCD, 0xDCBA],
                    },
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "advanced",
                        config.CONF_OTA_PROVIDER_PATH: tmp_path,
                        config.CONF_OTA_PROVIDER_WARNING: config.CONF_OTA_ALLOW_ADVANCED_DIR_STRING,
                    },
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "z2m_local",
                        config.CONF_OTA_PROVIDER_INDEX_FILE: tmp_path / "index.json",
                    },
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "zigpy_local",
                        config.CONF_OTA_PROVIDER_INDEX_FILE: tmp_path / "index.json",
                    },
                ],
            }
        ),
        application=None,
    )

    assert ota._providers == [
        # zigpy.ota.providers.ZigpyOtaProvider(),
        # zigpy.ota.providers.Ledvance(),
        # zigpy.ota.providers.Sonoff(),
        zigpy.ota.providers.Salus(url="https://salus.example.org/"),
        zigpy.ota.providers.RemoteZ2MProvider(url="https://z2m.example.org/"),
        zigpy.ota.providers.Tradfri(
            url="https://ikea3.example.org/",
            manufacturer_ids=[0xABCD, 0xDCBA],
        ),
        zigpy.ota.providers.AdvancedFileProvider(path=tmp_path),
        zigpy.ota.providers.LocalZ2MProvider(index_file=tmp_path / "index.json"),
        zigpy.ota.providers.LocalZigpyProvider(index_file=tmp_path / "index.json"),
    ]


async def test_ota_config_stub_providers() -> None:
    """Test YAML config loading when removed OTA providers are passed."""
    zigpy.ota.OTA(
        config=config.SCHEMA_OTA(
            {
                config.CONF_OTA_ENABLED: True,
                config.CONF_OTA_BROADCAST_ENABLED: False,
                config.CONF_OTA_EXTRA_PROVIDERS: [
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "salus",
                        config.CONF_OTA_PROVIDER_URL: "https://salus.example.org/",
                    },
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "thirdreality",
                        config.CONF_OTA_PROVIDER_URL: "https://thirdreality.example.org/",
                    },
                    {
                        config.CONF_OTA_PROVIDER_TYPE: "inovelli",
                        config.CONF_OTA_PROVIDER_URL: "https://inovelli.example.org/",
                    },
                ],
            }
        ),
        application=None,
    )


async def test_ota_broadcast_loop() -> None:
    app = make_app(
        {
            config.CONF_OTA: {
                config.CONF_OTA_ENABLED: True,
                config.CONF_OTA_BROADCAST_ENABLED: True,
                config.CONF_OTA_BROADCAST_INITIAL_DELAY: 0.1,
                config.CONF_OTA_BROADCAST_INTERVAL: 0.2,
            }
        }
    )

    with (
        patch.object(
            app.ota,
            "broadcast_notify",
            side_effect=[None, None, RuntimeError(), None, None, None],
        ) as mock_broadcast_notify,
        patch.object(app.ota, "check_all_devices_for_ota", return_value=None),
        patch("zigpy.ota.BROADCAST_SETTLE_DELAY", 0),
    ):
        await app.startup()
        assert app.ota._broadcast_loop_task is not None
        await asyncio.sleep(1)
        await app.shutdown()

    assert app.ota._broadcast_loop_task is None
    assert len(mock_broadcast_notify.mock_calls) == 5


async def test_ota_broadcast_loop_check_failure() -> None:
    app = make_app(
        {
            config.CONF_OTA: {
                config.CONF_OTA_ENABLED: True,
                config.CONF_OTA_BROADCAST_ENABLED: True,
                config.CONF_OTA_BROADCAST_INITIAL_DELAY: 0.1,
                config.CONF_OTA_BROADCAST_INTERVAL: 0.2,
            }
        }
    )

    with (
        patch.object(app.ota, "broadcast_notify", return_value=None),
        patch.object(
            app.ota,
            "check_all_devices_for_ota",
            side_effect=[RuntimeError("provider down"), None, None],
        ) as mock_check,
        patch("zigpy.ota.BROADCAST_SETTLE_DELAY", 0),
    ):
        await app.startup()
        await asyncio.sleep(0.5)
        await app.shutdown()

    # Loop continues despite the first check failing
    assert len(mock_check.mock_calls) >= 2


async def test_ota_broadcast() -> None:
    app = make_app({config.CONF_OTA: {config.CONF_OTA_ENABLED: True}})

    await app.startup()
    app.send_packet.reset_mock()
    await app.ota.broadcast_notify()
    await app.shutdown()

    assert len(app.send_packet.mock_calls) == 1
    assert app.send_packet.mock_calls[0].args[0].dst.addr_mode == t.AddrMode.Broadcast



"""Provide Event base classes for zigpy."""

from __future__ import annotations

import asyncio
from collections.abc import Callable, Generator
import contextlib
from contextvars import ContextVar
import dataclasses
from inspect import iscoroutinefunction
import logging
from typing import Any

_LOGGER = logging.getLogger(__package__)

_suppress_events: ContextVar[bool] = ContextVar("suppress_events", default=False)


@contextlib.contextmanager
def suppress_events() -> Generator[None, None, None]:
    """Context manager to suppress event emission."""
    token = _suppress_events.set(True)

    try:
        yield
    finally:
        _suppress_events.reset(token)


@dataclasses.dataclass(frozen=True, slots=True)
class EventListener:
    """Listener for an event."""

    callback: Callable
    with_context: bool


class EventBase:
    """Base class for event handling and emitting objects."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize event base."""
        super().__init__(*args, **kwargs)
        self._event_listeners: dict[str, list[EventListener]] = {}
        self._event_tasks: list[asyncio.Task] = []
        self._global_listeners: list[EventListener] = []

    def on_event(  # pylint: disable=invalid-name
        self, event_name: str, callback: Callable, with_context: bool = False
    ) -> Callable:
        """Register an event callback."""
        listener = EventListener(callback=callback, with_context=with_context)

        listeners: list = self._event_listeners.setdefault(event_name, [])
        listeners.append(listener)

        def unsubscribe() -> None:
            """Unsubscribe listeners."""
            if listener in listeners:
                listeners.remove(listener)

        return unsubscribe

    def on_all_events(  # pylint: disable=invalid-name
        self, callback: Callable, with_context: bool = False
    ) -> Callable:
        """Register a callback for all events."""
        listener = EventListener(callback=callback, with_context=with_context)
        self._global_listeners.append(listener)

        def unsubscribe() -> None:
            """Unsubscribe listeners."""
            if listener in self._global_listeners:
                self._global_listeners.remove(listener)

        return unsubscribe

    def once(
        self, event_name: str, callback: Callable, with_context: bool = False
    ) -> Callable:
        """Listen for an event exactly once."""
        if iscoroutinefunction(callback):

            async def async_event_listener(*args, **kwargs) -> None:
                unsub()
                task = asyncio.create_task(callback(*args, **kwargs))
                self._event_tasks.append(task)
                task.add_done_callback(self._event_tasks.remove)

            unsub = self.on_event(
                event_name, async_event_listener, with_context=with_context
            )
            return unsub  # noqa: RET504

        def event_listener(*args, **kwargs) -> None:
            unsub()
            callback(*args, **kwargs)

        unsub = self.on_event(event_name, event_listener, with_context=with_context)
        return unsub  # noqa: RET504

    def emit(self, event_name: str, data=None) -> None:
        """Run all callbacks for an event."""
        if _suppress_events.get():
            return

        listeners = [
            *self._event_listeners.get(event_name, []),
            *self._global_listeners,
        ]
        _LOGGER.debug(
            "Emitting event %s with data %r (%d listeners)",
            event_name,
            data,
            len(listeners),
        )

        for listener in listeners:
            if listener.with_context:
                call = listener.callback(event_name, data)
            else:
                call = listener.callback(data)

            if iscoroutinefunction(listener.callback):
                task = asyncio.create_task(call)
                self._event_tasks.append(task)
                task.add_done_callback(self._event_tasks.remove)

    def _handle_event_protocol(self, event) -> None:
        """Process an event based on event protocol."""
        _LOGGER.debug(
            "(%s) handling event protocol for event: %s", self.__class__.__name__, event
        )
        handler = getattr(self, f"handle_{event.event.replace(' ', '_')}", None)
        if handler is None:
            _LOGGER.warning("Received unknown event: %s", event)
            return
        if iscoroutinefunction(handler):
            task = asyncio.create_task(handler(event))
            self._event_tasks.append(task)
            task.add_done_callback(self._event_tasks.remove)
        else:
            handler(event)



"""Closures Functional Domain"""

from __future__ import annotations

from typing import Final

import zigpy.types as t
from zigpy.zcl import Cluster, foundation
from zigpy.zcl.foundation import (
    BaseAttributeDefs,
    BaseCommandDefs,
    ZCLAttributeDef,
    ZCLCommandDef,
)


class ShadeStatus(t.bitmap8):
    Operational = 0b00000001
    Adjusting = 0b00000010
    Opening = 0b00000100
    Motor_forward_is_opening = 0b00001000


class ShadeMode(t.enum8):
    Normal = 0x00
    Configure = 0x00
    Unknown = 0xFF


class Shade(Cluster):
    """Attributes and commands for configuring a shade"""

    ShadeStatus: Final = ShadeStatus
    ShadeMode: Final = ShadeMode

    cluster_id: Final[t.uint16_t] = 0x0100
    name: Final = "Shade Configuration"
    ep_attribute: Final = "shade"

    class AttributeDefs(BaseAttributeDefs):
        # Shade Information
        physical_closed_limit: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint16_t, access="r"
        )
        motor_step_size: Final = ZCLAttributeDef(id=0x0001, type=t.uint8_t, access="r")
        status: Final = ZCLAttributeDef(
            id=0x0002, type=ShadeStatus, access="rw", mandatory=True
        )
        # Shade Settings
        closed_limit: Final = ZCLAttributeDef(
            id=0x0010, type=t.uint16_t, access="rw", mandatory=True
        )
        mode: Final = ZCLAttributeDef(
            id=0x0012, type=ShadeMode, access="rw", mandatory=True
        )


class LockState(t.enum8):
    Not_fully_locked = 0x00
    Locked = 0x01
    Unlocked = 0x02
    Undefined = 0xFF


class LockType(t.enum8):
    Dead_bolt = 0x00
    Magnetic = 0x01
    Other = 0x02
    Mortise = 0x03
    Rim = 0x04
    Latch_bolt = 0x05
    Cylindrical_lock = 0x06
    Tubular_lock = 0x07
    Interconnected_lock = 0x08
    Dead_latch = 0x09
    Door_furniture = 0x0A


class DoorState(t.enum8):
    Open = 0x00
    Closed = 0x01
    Error_jammed = 0x02
    Error_forced_open = 0x03
    Error_unspecified = 0x04
    Undefined = 0xFF


class OperatingMode(t.enum8):
    Normal = 0x00
    Vacation = 0x01
    Privacy = 0x02
    No_RF_Lock_Unlock = 0x03
    Passage = 0x04


class SupportedOperatingModes(t.bitmap16):
    Normal = 0x0001
    Vacation = 0x0002
    Privacy = 0x0004
    No_RF = 0x0008
    Passage = 0x0010


class DefaultConfigurationRegister(t.bitmap16):
    Enable_Local_Programming = 0x0001
    Keypad_Interface_default_access = 0x0002
    RF_Interface_default_access = 0x0004
    Sound_Volume_non_zero = 0x0020
    Auto_Relock_time_non_zero = 0x0040
    Led_settings_non_zero = 0x0080


class ZigbeeSecurityLevel(t.enum8):
    Network_Security = 0x00
    APS_Security = 0x01


class AlarmMask(t.bitmap16):
    Deadbolt_Jammed = 0x0001
    Lock_Reset_to_Factory_Defaults = 0x0002
    Reserved = 0x0004
    RF_Module_Power_Cycled = 0x0008
    Tamper_Alarm_wrong_code_entry_limit = 0x0010
    Tamper_Alarm_front_escutcheon_removed = 0x0020
    Forced_Door_Open_under_Door_Lockec_Condition = 0x0040


class KeypadOperationEventMask(t.bitmap16):
    Manufacturer_specific = 0x0001
    Lock_source_keypad = 0x0002
    Unlock_source_keypad = 0x0004
    Lock_source_keypad_error_invalid_code = 0x0008
    Lock_source_keypad_error_invalid_schedule = 0x0010
    Unlock_source_keypad_error_invalid_code = 0x0020
    Unlock_source_keypad_error_invalid_schedule = 0x0040
    Non_Access_User_Operation = 0x0080


class RFOperationEventMask(t.bitmap16):
    Manufacturer_specific = 0x0001
    Lock_source_RF = 0x0002
    Unlock_source_RF = 0x0004
    Lock_source_RF_error_invalid_code = 0x0008
    Lock_source_RF_error_invalid_schedule = 0x0010
    Unlock_source_RF_error_invalid_code = 0x0020
    Unlock_source_RF_error_invalid_schedule = 0x0040


class ManualOperatitonEventMask(t.bitmap16):
    Manufacturer_specific = 0x0001
    Thumbturn_Lock = 0x0002
    Thumbturn_Unlock = 0x0004
    One_touch_lock = 0x0008
    Key_Lock = 0x0010
    Key_Unlock = 0x0020
    Auto_lock = 0x0040
    Schedule_Lock = 0x0080
    Schedule_Unlock = 0x0100
    Manual_Lock_key_or_thumbturn = 0x0200
    Manual_Unlock_key_or_thumbturn = 0x0400


class RFIDOperationEventMask(t.bitmap16):
    Manufacturer_specific = 0x0001
    Lock_source_RFID = 0x0002
    Unlock_source_RFID = 0x0004
    Lock_source_RFID_error_invalid_RFID_ID = 0x0008
    Lock_source_RFID_error_invalid_schedule = 0x0010
    Unlock_source_RFID_error_invalid_RFID_ID = 0x0020
    Unlock_source_RFID_error_invalid_schedule = 0x0040


class KeypadProgrammingEventMask(t.bitmap16):
    Manufacturer_Specific = 0x0001
    Master_code_changed = 0x0002
    PIN_added = 0x0004
    PIN_deleted = 0x0008
    PIN_changed = 0x0010


class RFProgrammingEventMask(t.bitmap16):
    Manufacturer_Specific = 0x0001
    PIN_added = 0x0004
    PIN_deleted = 0x0008
    PIN_changed = 0x0010
    RFID_code_added = 0x0020
    RFID_code_deleted = 0x0040


class RFIDProgrammingEventMask(t.bitmap16):
    Manufacturer_Specific = 0x0001
    RFID_code_added = 0x0020
    RFID_code_deleted = 0x0040


class OperationEventSource(t.enum8):
    Keypad = 0x00
    RF = 0x01
    Manual = 0x02
    RFID = 0x03
    Indeterminate = 0xFF


class OperationEvent(t.enum8):
    UnknownOrMfgSpecific = 0x00
    Lock = 0x01
    Unlock = 0x02
    LockFailureInvalidPINorID = 0x03
    LockFailureInvalidSchedule = 0x04
    UnlockFailureInvalidPINorID = 0x05
    UnlockFailureInvalidSchedule = 0x06
    OnTouchLock = 0x07
    KeyLock = 0x08
    KeyUnlock = 0x09
    AutoLock = 0x0A
    ScheduleLock = 0x0B
    ScheduleUnlock = 0x0C
    Manual_Lock = 0x0D
    Manual_Unlock = 0x0E
    Non_Access_User_Operational_Event = 0x0F


class ProgrammingEvent(t.enum8):
    UnknownOrMfgSpecific = 0x00
    MasterCodeChanged = 0x01
    PINCodeAdded = 0x02
    PINCodeDeleted = 0x03
    PINCodeChanges = 0x04
    RFIDCodeAdded = 0x05
    RFIDCodeDeleted = 0x06


class UserStatus(t.enum8):
    Available = 0x00
    Enabled = 0x01
    Disabled = 0x03
    Not_Supported = 0xFF


class UserType(t.enum8):
    Unrestricted = 0x00
    Year_Day_Schedule_User = 0x01
    Week_Day_Schedule_User = 0x02
    Master_User = 0x03
    Non_Access_User = 0x04
    Not_Supported = 0xFF


class DayMask(t.bitmap8):
    Sun = 0x01
    Mon = 0x02
    Tue = 0x04
    Wed = 0x08
    Thu = 0x10
    Fri = 0x20
    Sat = 0x40


class EventType(t.enum8):
    Operation = 0x00
    Programming = 0x01
    Alarm = 0x02


class DoorLock(Cluster):
    """The door lock cluster provides an interface to a generic way to secure a door."""

    LockState: Final = LockState
    LockType: Final = LockType
    DoorState: Final = DoorState
    OperatingMode: Final = OperatingMode
    SupportedOperatingModes: Final = SupportedOperatingModes
    DefaultConfigurationRegister: Final = DefaultConfigurationRegister
    ZigbeeSecurityLevel: Final = ZigbeeSecurityLevel
    AlarmMask: Final = AlarmMask
    KeypadOperationEventMask: Final = KeypadOperationEventMask
    RFOperationEventMask: Final = RFOperationEventMask
    ManualOperatitonEventMask: Final = ManualOperatitonEventMask
    RFIDOperationEventMask: Final = RFIDOperationEventMask
    KeypadProgrammingEventMask: Final = KeypadProgrammingEventMask
    RFProgrammingEventMask: Final = RFProgrammingEventMask
    RFIDProgrammingEventMask: Final = RFIDProgrammingEventMask
    OperationEventSource: Final = OperationEventSource
    OperationEvent: Final = OperationEvent
    ProgrammingEvent: Final = ProgrammingEvent
    UserStatus: Final = UserStatus
    UserType: Final = UserType
    DayMask: Final = DayMask
    EventType: Final = EventType

    cluster_id: Final[t.uint16_t] = 0x0101
    name: Final = "Door Lock"
    ep_attribute: Final = "door_lock"

    class AttributeDefs(BaseAttributeDefs):
        lock_state: Final = ZCLAttributeDef(
            id=0x0000, type=LockState, access="rp", mandatory=True
        )
        lock_type: Final = ZCLAttributeDef(
            id=0x0001, type=LockType, access="r", mandatory=True
        )
        actuator_enabled: Final = ZCLAttributeDef(
            id=0x0002, type=t.Bool, access="r", mandatory=True
        )
        door_state: Final = ZCLAttributeDef(id=0x0003, type=DoorState, access="rp")
        door_open_events: Final = ZCLAttributeDef(
            id=0x0004, type=t.uint32_t, access="rw"
        )
        door_closed_events: Final = ZCLAttributeDef(
            id=0x0005, type=t.uint32_t, access="rw"
        )
        open_period: Final = ZCLAttributeDef(id=0x0006, type=t.uint16_t, access="rw")
        num_of_lock_records_supported: Final = ZCLAttributeDef(
            id=0x0010, type=t.uint16_t, access="r"
        )
        num_of_total_users_supported: Final = ZCLAttributeDef(
            id=0x0011, type=t.uint16_t, access="r"
        )
        num_of_pin_users_supported: Final = ZCLAttributeDef(
            id=0x0012, type=t.uint16_t, access="r"
        )
        num_of_rfid_users_supported: Final = ZCLAttributeDef(
            id=0x0013, type=t.uint16_t, access="r"
        )
        num_of_week_day_schedules_supported_per_user: Final = ZCLAttributeDef(
            id=0x0014, type=t.uint8_t, access="r"
        )
        num_of_year_day_schedules_supported_per_user: Final = ZCLAttributeDef(
            id=0x0015, type=t.uint8_t, access="r"
        )
        num_of_holiday_scheduleds_supported: Final = ZCLAttributeDef(
            id=0x0016, type=t.uint8_t, access="r"
        )
        max_pin_len: Final = ZCLAttributeDef(id=0x0017, type=t.uint8_t, access="r")
        min_pin_len: Final = ZCLAttributeDef(id=0x0018, type=t.uint8_t, access="r")
        max_rfid_len: Final = ZCLAttributeDef(id=0x0019, type=t.uint8_t, access="r")
        min_rfid_len: Final = ZCLAttributeDef(id=0x001A, type=t.uint8_t, access="r")
        enable_logging: Final = ZCLAttributeDef(id=0x0020, type=t.Bool, access="r*wp")
        language: Final = ZCLAttributeDef(
            id=0x0021, type=t.LimitedCharString(3), access="r*wp"
        )
        led_settings: Final = ZCLAttributeDef(id=0x0022, type=t.uint8_t, access="r*wp")
        auto_relock_time: Final = ZCLAttributeDef(
            id=0x0023, type=t.uint32_t, access="r*wp"
        )
        sound_volume: Final = ZCLAttributeDef(id=0x0024, type=t.uint8_t, access="r*wp")
        operating_mode: Final = ZCLAttributeDef(
            id=0x0025, type=OperatingMode, access="r*wp"
        )
        supported_operating_modes: Final = ZCLAttributeDef(
            id=0x0026, type=SupportedOperatingModes, access="r"
        )
        default_configuration_register: Final = ZCLAttributeDef(
            id=0x0027,
            type=DefaultConfigurationRegister,
            access="rp",
        )
        enable_local_programming: Final = ZCLAttributeDef(
            id=0x0028, type=t.Bool, access="r*wp"
        )
        enable_one_touch_locking: Final = ZCLAttributeDef(
            id=0x0029, type=t.Bool, access="rwp"
        )
        enable_inside_status_led: Final = ZCLAttributeDef(
            id=0x002A, type=t.Bool, access="rwp"
        )
        enable_privacy_mode_button: Final = ZCLAttributeDef(
            id=0x002B, type=t.Bool, access="rwp"
        )
        wrong_code_entry_limit: Final = ZCLAttributeDef(
            id=0x0030, type=t.uint8_t, access="r*wp"
        )
        user_code_temporary_disable_time: Final = ZCLAttributeDef(
            id=0x0031, type=t.uint8_t, access="r*wp"
        )
        send_pin_ota: Final = ZCLAttributeDef(id=0x0032, type=t.Bool, access="r*wp")
        require_pin_for_rf_operation: Final = ZCLAttributeDef(
            id=0x0033, type=t.Bool, access="r*wp"
        )
        zigbee_security_level: Final = ZCLAttributeDef(
            id=0x0034, type=ZigbeeSecurityLevel, access="rp"
        )
        alarm_mask: Final = ZCLAttributeDef(id=0x0040, type=AlarmMask, access="rwp")
        keypad_operation_event_mask: Final = ZCLAttributeDef(
            id=0x0041, type=KeypadOperationEventMask, access="rwp"
        )
        rf_operation_event_mask: Final = ZCLAttributeDef(
            id=0x0042, type=RFOperationEventMask, access="rwp"
        )
        manual_operation_event_mask: Final = ZCLAttributeDef(
            id=0x0043, type=ManualOperatitonEventMask, access="rwp"
        )
        rfid_operation_event_mask: Final = ZCLAttributeDef(
            id=0x0044, type=RFIDOperationEventMask, access="rwp"
        )
        keypad_programming_event_mask: Final = ZCLAttributeDef(
            id=0x0045,
            type=KeypadProgrammingEventMask,
            access="rwp",
        )
        rf_programming_event_mask: Final = ZCLAttributeDef(
            id=0x0046, type=RFProgrammingEventMask, access="rwp"
        )
        rfid_programming_event_mask: Final = ZCLAttributeDef(
            id=0x0047, type=RFIDProgrammingEventMask, access="rwp"
        )

    class ServerCommandDefs(BaseCommandDefs):
        lock_door: Final = ZCLCommandDef(
            id=0x00,
            schema={"pin_code?": t.CharacterString},
        )
        unlock_door: Final = ZCLCommandDef(
            id=0x01,
            schema={"pin_code?": t.CharacterString},
        )
        toggle_door: Final = ZCLCommandDef(
            id=0x02,
            schema={"pin_code?": t.CharacterString},
        )
        unlock_with_timeout: Final = ZCLCommandDef(
            id=0x03,
            schema={"timeout": t.uint16_t, "pin_code?": t.CharacterString},
        )
        get_log_record: Final = ZCLCommandDef(
            id=0x04,
            schema={"log_index": t.uint16_t},
        )
        set_pin_code: Final = ZCLCommandDef(
            id=0x05,
            schema={
                "user_id": t.uint16_t,
                "user_status": UserStatus,
                "user_type": UserType,
                "pin_code": t.CharacterString,
            },
        )
        get_pin_code: Final = ZCLCommandDef(
            id=0x06,
            schema={"user_id": t.uint16_t},
        )
        clear_pin_code: Final = ZCLCommandDef(
            id=0x07,
            schema={"user_id": t.uint16_t},
        )
        clear_all_pin_codes: Final = ZCLCommandDef(id=0x08, schema={})
        set_user_status: Final = ZCLCommandDef(
            id=0x09,
            schema={"user_id": t.uint16_t, "user_status": UserStatus},
        )
        get_user_status: Final = ZCLCommandDef(
            id=0x0A,
            schema={"user_id": t.uint16_t},
        )
        set_week_day_schedule: Final = ZCLCommandDef(
            id=0x0B,
            schema={
                "schedule_id": t.uint8_t,
                "user_id": t.uint16_t,
                "days_mask": DayMask,
                "start_hour": t.uint8_t,
                "start_minute": t.uint8_t,
                "end_hour": t.uint8_t,
                "end_minute": t.uint8_t,
            },
        )
        get_week_day_schedule: Final = ZCLCommandDef(
            id=0x0C,
            schema={"schedule_id": t.uint8_t, "user_id": t.uint16_t},
        )
        clear_week_day_schedule: Final = ZCLCommandDef(
            id=0x0D,
            schema={"schedule_id": t.uint8_t, "user_id": t.uint16_t},
        )
        set_year_day_schedule: Final = ZCLCommandDef(
            id=0x0E,
            schema={
                "schedule_id": t.uint8_t,
                "user_id": t.uint16_t,
                "local_start_time": t.LocalTime,
                "local_end_time": t.LocalTime,
            },
        )
        get_year_day_schedule: Final = ZCLCommandDef(
            id=0x0F,
            schema={"schedule_id": t.uint8_t, "user_id": t.uint16_t},
        )
        clear_year_day_schedule: Final = ZCLCommandDef(
            id=0x10,
            schema={"schedule_id": t.uint8_t, "user_id": t.uint16_t},
        )
        set_holiday_schedule: Final = ZCLCommandDef(
            id=0x11,
            schema={
                "holiday_schedule_id": t.uint8_t,
                "local_start_time": t.LocalTime,
                "local_end_time": t.LocalTime,
                "operating_mode_during_holiday": OperatingMode,
            },
        )
        get_holiday_schedule: Final = ZCLCommandDef(
            id=0x12,
            schema={"holiday_schedule_id": t.uint8_t},
        )
        clear_holiday_schedule: Final = ZCLCommandDef(
            id=0x13,
            schema={"holiday_schedule_id": t.uint8_t},
        )
        set_user_type: Final = ZCLCommandDef(
            id=0x14,
            schema={"user_id": t.uint16_t, "user_type": UserType},
        )
        get_user_type: Final = ZCLCommandDef(
            id=0x15,
            schema={"user_id": t.uint16_t},
        )
        set_rfid_code: Final = ZCLCommandDef(
            id=0x16,
            schema={
                "user_id": t.uint16_t,
                "user_status": UserStatus,
                "user_type": UserType,
                "rfid_code": t.CharacterString,
            },
        )
        get_rfid_code: Final = ZCLCommandDef(
            id=0x17,
            schema={"user_id": t.uint16_t},
        )
        clear_rfid_code: Final = ZCLCommandDef(
            id=0x18,
            schema={"user_id": t.uint16_t},
        )
        clear_all_rfid_codes: Final = ZCLCommandDef(id=0x19, schema={})

    class ClientCommandDefs(BaseCommandDefs):
        lock_door_response: Final = ZCLCommandDef(
            id=0x00,
            schema={"status": foundation.Status},
        )
        unlock_door_response: Final = ZCLCommandDef(
            id=0x01,
            schema={"status": foundation.Status},
        )
        toggle_door_response: Final = ZCLCommandDef(
            id=0x02,
            schema={"status": foundation.Status},
        )
        unlock_with_timeout_response: Final = ZCLCommandDef(
            id=0x03,
            schema={"status": foundation.Status},
        )
        get_log_record_response: Final = ZCLCommandDef(
            id=0x04,
            schema={
                "log_entry_id": t.uint16_t,
                "timestamp": t.uint32_t,
                "event_type": EventType,
                "source": OperationEventSource,
                "event_id_or_alarm_code": t.uint8_t,
                "user_id": t.uint16_t,
                "pin?": t.CharacterString,
            },
        )
        set_pin_code_response: Final = ZCLCommandDef(
            id=0x05,
            schema={"status": foundation.Status},
        )
        get_pin_code_response: Final = ZCLCommandDef(
            id=0x06,
            schema={
                "user_id": t.uint16_t,
                "user_status": UserStatus,
                "user_type": UserType,
                "code": t.CharacterString,
            },
        )
        clear_pin_code_response: Final = ZCLCommandDef(
            id=0x07,
            schema={"status": foundation.Status},
        )
        clear_all_pin_codes_response: Final = ZCLCommandDef(
            id=0x08,
            schema={"status": foundation.Status},
        )
        set_user_status_response: Final = ZCLCommandDef(
            id=0x09,
            schema={"status": foundation.Status},
        )
        get_user_status_response: Final = ZCLCommandDef(
            id=0x0A,
            schema={"user_id": t.uint16_t, "user_status": UserStatus},
        )
        set_week_day_schedule_response: Final = ZCLCommandDef(
            id=0x0B,
            schema={"status": foundation.Status},
        )
        get_week_day_schedule_response: Final = ZCLCommandDef(
            id=0x0C,
            schema={
                "schedule_id": t.uint8_t,
                "user_id": t.uint16_t,
                "status": foundation.Status,
                "days_mask?": t.uint8_t,
                "start_hour?": t.uint8_t,
                "start_minute?": t.uint8_t,
                "end_hour?": t.uint8_t,
                "end_minute?": t.uint8_t,
            },
        )
        clear_week_day_schedule_response: Final = ZCLCommandDef(
            id=0x0D,
            schema={"status": foundation.Status},
        )
        set_year_day_schedule_response: Final = ZCLCommandDef(
            id=0x0E,
            schema={"status": foundation.Status},
        )
        get_year_day_schedule_response: Final = ZCLCommandDef(
            id=0x0F,
            schema={
                "schedule_id": t.uint8_t,
                "user_id": t.uint16_t,
                "status": foundation.Status,
                "local_start_time?": t.LocalTime,
                "local_end_time?": t.LocalTime,
            },
        )
        clear_year_day_schedule_response: Final = ZCLCommandDef(
            id=0x10,
            schema={"status": foundation.Status},
        )
        set_holiday_schedule_response: Final = ZCLCommandDef(
            id=0x11,
            schema={"status": foundation.Status},
        )
        get_holiday_schedule_response: Final = ZCLCommandDef(
            id=0x12,
            schema={
                "holiday_schedule_id": t.uint8_t,
                "status": foundation.Status,
                "local_start_time?": t.LocalTime,
                "local_end_time?": t.LocalTime,
                "operating_mode_during_holiday?": t.uint8_t,
            },
        )
        clear_holiday_schedule_response: Final = ZCLCommandDef(
            id=0x13,
            schema={"status": foundation.Status},
        )
        set_user_type_response: Final = ZCLCommandDef(
            id=0x14,
            schema={"status": foundation.Status},
        )
        get_user_type_response: Final = ZCLCommandDef(
            id=0x15,
            schema={"user_id": t.uint16_t, "user_type": UserType},
        )
        set_rfid_code_response: Final = ZCLCommandDef(
            id=0x16,
            schema={"status": foundation.Status},
        )
        get_rfid_code_response: Final = ZCLCommandDef(
            id=0x17,
            schema={
                "user_id": t.uint16_t,
                "user_status": UserStatus,
                "user_type": UserType,
                "rfid_code": t.CharacterString,
            },
        )
        clear_rfid_code_response: Final = ZCLCommandDef(
            id=0x18,
            schema={"status": foundation.Status},
        )
        clear_all_rfid_codes_response: Final = ZCLCommandDef(
            id=0x19,
            schema={"status": foundation.Status},
        )
        operation_event_notification: Final = ZCLCommandDef(
            id=0x20,
            schema={
                "operation_event_source": OperationEventSource,
                "operation_event_code": OperationEvent,
                "user_id": t.uint16_t,
                "pin": t.CharacterString,
                "local_time": t.LocalTime,
                "data?": t.CharacterString,
            },
        )
        programming_event_notification: Final = ZCLCommandDef(
            id=0x21,
            schema={
                "program_event_source": OperationEventSource,
                "program_event_code": ProgrammingEvent,
                "user_id": t.uint16_t,
                "pin": t.CharacterString,
                "user_type": UserType,
                "user_status": UserStatus,
                "local_time": t.LocalTime,
                "data?": t.CharacterString,
            },
        )


class WindowCoveringType(t.enum8):
    Rollershade = 0x00
    Rollershade_two_motors = 0x01
    Rollershade_exterior = 0x02
    Rollershade_exterior_two_motors = 0x03
    Drapery = 0x04
    Awning = 0x05
    Shutter = 0x06
    Tilt_blind_tilt_only = 0x07
    Tilt_blind_tilt_and_lift = 0x08
    Projector_screen = 0x09


class ConfigStatus(t.bitmap8):
    Operational = 0b00000001
    Online = 0b00000010
    Open_up_commands_reversed = 0b00000100
    Closed_loop_lift_control = 0b00001000
    Closed_loop_tilt_control = 0b00010000
    Encoder_controlled_lift = 0b00100000
    Encoder_controlled_tilt = 0b01000000


class WindowCoveringMode(t.bitmap8):
    Motor_direction_reversed = 0b00000001
    Run_in_calibration_mode = 0b00000010
    Motor_in_maintenance_mode = 0b00000100
    LEDs_display_feedback = 0b00001000


class WindowCovering(Cluster):
    WindowCoveringType: Final = WindowCoveringType
    ConfigStatus: Final = ConfigStatus
    WindowCoveringMode: Final = WindowCoveringMode

    cluster_id: Final[t.uint16_t] = 0x0102
    name: Final = "Window Covering"
    ep_attribute: Final = "window_covering"

    class AttributeDefs(BaseAttributeDefs):
        # Window Covering Information
        window_covering_type: Final = ZCLAttributeDef(
            id=0x0000, type=WindowCoveringType, access="r", mandatory=True
        )
        physical_closed_limit_lift: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint16_t, access="r"
        )
        physical_closed_limit_tilt: Final = ZCLAttributeDef(
            id=0x0002, type=t.uint16_t, access="r"
        )
        current_position_lift: Final = ZCLAttributeDef(
            id=0x0003, type=t.uint16_t, access="r"
        )
        current_position_tilt: Final = ZCLAttributeDef(
            id=0x0004, type=t.uint16_t, access="r"
        )
        number_of_actuations_lift: Final = ZCLAttributeDef(
            id=0x0005, type=t.uint16_t, access="r"
        )
        number_of_actuations_tilt: Final = ZCLAttributeDef(
            id=0x0006, type=t.uint16_t, access="r"
        )
        config_status: Final = ZCLAttributeDef(
            id=0x0007, type=ConfigStatus, access="r", mandatory=True
        )
        # All subsequent attributes are mandatory if their control types are enabled
        current_position_lift_percentage: Final = ZCLAttributeDef(
            id=0x0008, type=t.uint8_t, access="rps"
        )
        current_position_tilt_percentage: Final = ZCLAttributeDef(
            id=0x0009, type=t.uint8_t, access="rps"
        )
        # Window Covering Settings
        installed_open_limit_lift: Final = ZCLAttributeDef(
            id=0x0010, type=t.uint16_t, access="r"
        )
        installed_closed_limit_lift: Final = ZCLAttributeDef(
            id=0x0011, type=t.uint16_t, access="r"
        )
        installed_open_limit_tilt: Final = ZCLAttributeDef(
            id=0x0012, type=t.uint16_t, access="r"
        )
        installed_closed_limit_tilt: Final = ZCLAttributeDef(
            id=0x0013, type=t.uint16_t, access="r"
        )
        velocity_lift: Final = ZCLAttributeDef(id=0x0014, type=t.uint16_t, access="rw")
        acceleration_time_lift: Final = ZCLAttributeDef(
            id=0x0015, type=t.uint16_t, access="rw"
        )
        deceleration_time_lift: Final = ZCLAttributeDef(
            id=0x0016, type=t.uint16_t, access="rw"
        )
        window_covering_mode: Final = ZCLAttributeDef(
            id=0x0017, type=WindowCoveringMode, access="rw", mandatory=True
        )
        intermediate_setpoints_lift: Final = ZCLAttributeDef(
            id=0x0018, type=t.LVBytes, access="rw"
        )
        intermediate_setpoints_tilt: Final = ZCLAttributeDef(
            id=0x0019, type=t.LVBytes, access="rw"
        )

    class ServerCommandDefs(BaseCommandDefs):
        up_open: Final = ZCLCommandDef(id=0x00, schema={})
        down_close: Final = ZCLCommandDef(id=0x01, schema={})
        stop: Final = ZCLCommandDef(id=0x02, schema={})
        go_to_lift_value: Final = ZCLCommandDef(
            id=0x04,
            schema={"lift_value": t.uint16_t},
        )
        go_to_lift_percentage: Final = ZCLCommandDef(
            id=0x05,
            schema={"percentage_lift_value": t.uint8_t},
        )
        go_to_tilt_value: Final = ZCLCommandDef(
            id=0x07,
            schema={"tilt_value": t.uint16_t},
        )
        go_to_tilt_percentage: Final = ZCLCommandDef(
            id=0x08,
            schema={"percentage_tilt_value": t.uint8_t},
        )


class MovingState(t.enum8):
    Stopped = 0x00
    Closing = 0x01
    Opening = 0x02


class SafetyStatus(t.bitmap16):
    Remote_Lockout = 0b00000000_00000001
    Tamper_Detected = 0b00000000_00000010
    Failed_Communication = 0b00000000_00000100
    Position_Failure = 0b00000000_00001000


class Capabilities(t.bitmap8):
    Partial_Barrier = 0b00000001


class BarrierControl(Cluster):
    cluster_id: Final = 0x0103
    name: Final = "Barrier Control"
    ep_attribute: Final = "barrier_control"

    class AttributeDefs(BaseAttributeDefs):
        moving_state: Final = ZCLAttributeDef(
            id=0x0001, type=MovingState, access="rp", mandatory=True
        )
        safety_status: Final = ZCLAttributeDef(
            id=0x0002, type=SafetyStatus, access="rp", mandatory=True
        )
        capabilities: Final = ZCLAttributeDef(
            id=0x0003, type=Capabilities, access="r", mandatory=True
        )
        open_events: Final = ZCLAttributeDef(id=0x0004, type=t.uint16_t, access="rw")
        close_events: Final = ZCLAttributeDef(id=0x0005, type=t.uint16_t, access="rw")
        command_open_events: Final = ZCLAttributeDef(
            id=0x0006, type=t.uint16_t, access="rw"
        )
        command_close_events: Final = ZCLAttributeDef(
            id=0x0007, type=t.uint16_t, access="rw"
        )
        open_period: Final = ZCLAttributeDef(id=0x0008, type=t.uint16_t, access="rw")
        close_period: Final = ZCLAttributeDef(id=0x0009, type=t.uint16_t, access="rw")
        barrier_position: Final = ZCLAttributeDef(
            id=0x000A, type=t.uint8_t, access="rps", mandatory=True
        )

    class ServerCommandDefs(BaseCommandDefs):
        go_to_percent: Final = ZCLCommandDef(
            id=0x00,
            schema={"percent_open": t.uint8_t},
        )
        stop: Final = ZCLCommandDef(id=0x01, schema={})

    class ClientCommandDefs(BaseCommandDefs):
        pass
